from contextvars import ContextVar, Token
from logging import Logger

from fastapi import BackgroundTasks
from kink import di
from libre_fastapi_jwt import AuthJWT

from appodus_utils.domain.user.auth.active_auditor.models import ActiveAuditor
from appodus_utils.exception.exceptions import NoActiveSessionException

logger: Logger = di['logger']
# Context variable to hold BackgroundTasks, ActiveAuditor and AuthJWT per context
background_tasks_ctx: ContextVar[BackgroundTasks] = ContextVar("background_tasks_ctx")
active_auditor_ctx: ContextVar[ActiveAuditor] = ContextVar("active_auditor_ctx")
authorizer_ctx: ContextVar[AuthJWT] = ContextVar("authorizer_ctx")


class ActiveAuditorContext:

    # ActiveAuditor
    @staticmethod
    def set_active_auditor_context(active_auditor: ActiveAuditor) -> Token[ActiveAuditor]:
        logger.debug(f"Setting active auditor in context: {active_auditor}")
        return active_auditor_ctx.set(active_auditor)

    @staticmethod
    def reset_active_auditor_context(token: Token[ActiveAuditor]):
        return active_auditor_ctx.reset(token)

    @staticmethod
    def get_active_auditor_from_context() -> ActiveAuditor:
        error_msg = (
            "No active_auditor found in context. "
            "Ensure you're within a request/task context that has set this value."
        )
        try:
            active_auditor = active_auditor_ctx.get()
            if not active_auditor:
                logger.error("active_auditor_ctx contains None.")
                raise NoActiveSessionException(error_msg)
            logger.debug(f"Retrieved active auditor from context: {active_auditor}")
            return active_auditor
        except LookupError:
            logger.exception("active_auditor_ctx was not set.")
            raise NoActiveSessionException(error_msg)

    # AuthJWT
    @staticmethod
    def set_authorizer_context(authorizer: AuthJWT) -> Token[AuthJWT]:
        logger.debug(f"Setting active auditor in context: {authorizer}")
        return authorizer_ctx.set(authorizer)

    @staticmethod
    def reset_authorizer_context(token: Token[AuthJWT]):
        return authorizer_ctx.reset(token)

    @staticmethod
    def get_authorizer_from_context() -> AuthJWT:
        error_msg = (
            "No authorizer found in context. "
            "Ensure you're within a request/task context that has set this value."
        )
        try:
            authorizer = authorizer_ctx.get()
            if not authorizer:
                logger.error("authorizer_ctx contains None.")
                raise NoActiveSessionException(error_msg)
            logger.debug(f"Retrieved active auditor from context: {authorizer}")
            return authorizer
        except NoActiveSessionException:
            logger.exception("authorizer_ctx was not set.")
            raise RuntimeError(error_msg)

    # BackgroundTasks
    @staticmethod
    def set_background_tasks_context(background_tasks: BackgroundTasks) -> Token[BackgroundTasks]:
        logger.debug(f"Setting background_tasks in context: {background_tasks}")
        return background_tasks_ctx.set(background_tasks)

    @staticmethod
    def reset_background_tasks_context(token: Token[BackgroundTasks]):
        return background_tasks_ctx.reset(token)

    @staticmethod
    def get_background_tasks_from_context() -> BackgroundTasks:
        error_msg = (
            "No background_tasks found in context. "
            "Ensure you're within a request/task context that has set this value."
        )
        try:
            background_tasks = background_tasks_ctx.get()
            if not background_tasks:
                logger.error("background_tasks_ctx contains None.")
                raise NoActiveSessionException(error_msg)
            logger.debug(f"Retrieved background_tasks from context: {background_tasks}")
            return background_tasks
        except LookupError:
            logger.exception("background_tasks_ctx was not set.")
            raise NoActiveSessionException(error_msg)



from typing import Optional

from sqlalchemy import Column, String, JSON, Boolean
from sqlalchemy.ext.mutable import MutableDict

from appodus_utils import BaseEntity, PageRequest, BaseQueryDto, Object
from appodus_utils.db.types.loci import Loci


class Address(BaseEntity):
    __tablename__ = 'addresses'
    user_id = Column(String(36), nullable=False, index=True, unique=False)
    street_number = Column(String(10), nullable=False)
    street_name = Column(String(40), nullable=False)
    neighborhood = Column(String(20), nullable=False)
    locality = Column(String(20), nullable=False)
    lga = Column(String(36), nullable=False)
    state = Column(String(36), nullable=False)
    country_id = Column(String(36), nullable=False)
    gps_location = Column(MutableDict.as_mutable(JSON), nullable=False)
    verified = Column(Boolean, default=False, nullable=False)


class AddressBaseDto(Object):
    street_number: str
    street_name: str
    neighborhood: str
    locality: str
    lga: str
    state: str
    country_id: str
    gps_location: Loci


class CreateAddressDto(AddressBaseDto):
    user_id: str


class UpdateAddressDto(AddressBaseDto):
    pass


class _UpdateAddressDto(Object):
    verified: Optional[bool] = None


class SearchAddressDto(PageRequest, BaseQueryDto):
    user_id: Optional[str] = None
    street_number: Optional[str] = None
    street_name: Optional[str] = None
    neighborhood: Optional[str] = None
    locality: Optional[str] = None
    lga: Optional[str] = None
    state: Optional[str] = None
    country_id: Optional[str] = None
    verified: Optional[bool] = None


class QueryAddressDto(AddressBaseDto, _UpdateAddressDto, BaseQueryDto):
    user_id: str



import enum
from decimal import Decimal

from pydantic import Field, ConfigDict

from appodus_utils import Object


class TransactionCurrency(str, enum.Enum):
    USD = 'USD'
    EUR = 'EUR'
    GBP = 'GBP'
    NGN = 'NGN'


class Money(Object):
    value: Decimal = Field(0.0)
    currency: TransactionCurrency = Field(None)

    def get_value(self) -> Decimal:
        return self.value

    def get_currency(self) -> TransactionCurrency:
        return self.currency

    def plus(self, money: 'Money') -> 'Money':
        self._validate_currency(money.get_currency())
        self._validate_negative_credit(money.get_value())
        self.value += money.get_value()

        return self

    def minus(self, money: 'Money') -> 'Money':
        self._validate_debit_balance(money.get_value())
        self._validate_currency(money.get_currency())
        self._validate_negative_debit(money.get_value())
        self.value -= money.get_value()

        return self

    def compare(self, other: 'Money') -> Decimal:
        """
        Compare self to other.  Return a decimal value:

            a or b is a NaN ==> Decimal('NaN')
            a < b           ==> Decimal('-1')
            a == b          ==> Decimal('0')
            a > b           ==> Decimal('1')

        :param other:
        :return:
        """
        self._validate_currency(other.get_currency())
        return self.value.compare(other.get_value())

    def is_less_than(self, other: 'Money') -> bool:
        return self.value.compare(other.get_value()) < 0

    def is_greater_than(self, other: 'Money') -> bool:
        return self.value.compare(other.get_value()) > 0

    def is_equal_to(self, other: 'Money') -> bool:
        return self.value.compare(other.get_value()) == 0

    def _validate_currency(self, currency: TransactionCurrency):
        if self.currency != currency:
            raise ValueError("Money exception: currency mismatch")

    @staticmethod
    def _validate_negative_credit(amount: Decimal):
        if amount.compare(Decimal(0)) == Decimal(-1):
            raise ValueError("Money exception: negative credit amount")

    @staticmethod
    def _validate_negative_debit(amount: Decimal):
        if amount.compare(Decimal(0)) == Decimal(-1):
            raise ValueError("Money exception: negative debit amount")

    def _validate_debit_balance(self, amount: Decimal):
        if self.value.compare(Decimal(0)) == Decimal(-1) or self.value.compare(amount) == Decimal(-1):
            raise ValueError("Money exception: insufficient balance")

    model_config = ConfigDict(
        # strict=True,
        populate_by_name=True,  # allows using the alias when parsing
        # extra="forbid"          # disallow any extra fields not defined here
    )



from appodus_utils import Page, Utils
from fastapi import APIRouter, Depends, status
from fastapi.security import OAuth2PasswordRequestForm
from kink import di

from appodus_utils.db.models import SuccessResponse
from appodus_utils.domain.user.auth.models import EmailLoginRequest, Token, LoginSuccessDto, ChangePasswordDto, \
    PasswordResetRequest, PasswordResetConfirm
from appodus_utils.domain.user.auth.service import AuthService
from appodus_utils.domain.user.auth.social_login.controller import social_auth_router
from appodus_utils.domain.user.models import QueryUserDto, SearchUserDto
from appodus_utils.domain.user.service import UserService

auth_path = Utils.get_from_env_fail_if_not_exists("AUTH_URL_PATH")
auth_router = APIRouter(prefix=auth_path, tags=["Auth"])
auth_router.include_router(social_auth_router)

auth_service = di[AuthService]
user_service = di[UserService]


@auth_router.post("/access-token", summary='Authenticate and get access token', response_model=Token,
                  status_code=status.HTTP_200_OK)
async def access_token(auth_data: OAuth2PasswordRequestForm = Depends()) -> Token:
    login_dto = EmailLoginRequest(username=auth_data.username, password=auth_data.password)
    return (await auth_service.basic_login(login_dto)).token


@auth_router.post("/login", summary='Login', response_model=LoginSuccessDto, status_code=status.HTTP_200_OK)
async def login(auth_data: EmailLoginRequest) -> LoginSuccessDto:
    return await auth_service.basic_login(auth_data)


@auth_router.post("/refresh-token", response_model=SuccessResponse[bool])
async def auth_refresh_token() -> SuccessResponse[bool]:
    return await auth_service.refresh_token()


@auth_router.post("/logout", response_model=SuccessResponse[bool])
async def auth_logout() -> SuccessResponse[bool]:
    return await auth_service.logout()


@auth_router.get("/logged-users", response_model=Page[QueryUserDto])
async def auth_logged_users(search_dto: SearchUserDto):
    return await user_service.get_logged_user_page(search_dto)


@auth_router.patch("/change-password", response_model=SuccessResponse[bool])
async def change_password(payload: ChangePasswordDto) -> SuccessResponse[bool]:
    return await auth_service.change_password(payload)


@auth_router.post("/recover-password/message", response_model=SuccessResponse[bool])
async def send_password_recovery_message(payload: PasswordResetRequest) -> SuccessResponse[bool]:
    return await auth_service.send_password_recovery_message(payload)


@auth_router.post("/recover-password", response_model=LoginSuccessDto)
async def recover_password(payload: PasswordResetConfirm):
    return await auth_service.recover_password(payload)


@auth_router.get("/profile", response_model=LoginSuccessDto)
async def get_active_user_profile():
    return {}



from kink import inject

from appodus_utils.domain.user.repo import UserRepo
from appodus_utils.exception.exceptions import ResourceNotFoundException


@inject
class AuthValidator:
    def __init__(self, user_repo: UserRepo):
        self._user_repo = user_repo

    async def should_exist_by_id(self, _id: str):
        if not (await self._user_repo.exists_by_id(_id)):
            raise ResourceNotFoundException("User")



import base64
import hashlib
import secrets
from datetime import timedelta

from appodus_utils import Utils
from fastapi.params import Depends
from fastapi.security import OAuth2PasswordBearer
from libre_fastapi_jwt import AuthJWT

from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.db.redis_utils import RedisUtils
from appodus_utils.domain.user.auth.active_auditor.models import ActiveAuditor
from appodus_utils.domain.user.auth.active_auditor.service import ActiveAuditorService
from appodus_utils.domain.user.auth.models import TokenType, LoginSuccessDto

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="auths/access-token", auto_error=False)

# appodus_settings = Utils.get_from_env_fail_if_not_exists('APPODUS_SETTINGS')
# appodus_settings_dict = json.loads(appodus_settings)
# 
# utils_settings.copy(update=appodus_settings_dict)


class JwtAuthUtils:
    @staticmethod
    @AuthJWT.load_config
    def get_config():
        return utils_settings

    @staticmethod
    @AuthJWT.token_in_denylist_loader
    async def check_if_token_in_denylist(decrypted_token) -> bool:
        jti = decrypted_token['jti']
        entry = await RedisUtils.get_redis(jti)
        return entry and entry == 'true'

    @staticmethod
    async def revoke_token(authorizer: AuthJWT, token_type: TokenType = TokenType.ACCESS) -> bool:
        authorizer.jwt_required()
        jti = authorizer.get_raw_jwt()['jti']
        time_to_live = timedelta(
            seconds=utils_settings.ACCESS_TOKEN_EXPIRE_SECONDS) if token_type is TokenType.ACCESS else timedelta(
            seconds=utils_settings.REFRESH_TOKEN_EXPIRE_SECONDS)

        await RedisUtils.set_redis(jti, 'true', time_to_live)  # Store until token expires

        authorizer.unset_jwt_cookies()

        return True

    @staticmethod
    def set_access_token(user: ActiveAuditor, authorizer: AuthJWT):
        subject = user.id
        access_token_expires = timedelta(seconds=utils_settings.ACCESS_TOKEN_EXPIRE_SECONDS)
        refresh_token_expires = timedelta(seconds=utils_settings.REFRESH_TOKEN_EXPIRE_SECONDS)
        add_claims = {'ip': '127.0.0.1', 'user_type': user.user_type}  # TODO populate with user roles
        access_token = authorizer.create_access_token(subject=subject, expires_time=access_token_expires,
            user_claims=add_claims)
        refresh_token = authorizer.create_refresh_token(subject=subject, expires_time=refresh_token_expires,
            user_claims=add_claims)

        # Set the JWT and CSRF double submit cookies in the response
        try:
            authorizer.set_access_cookies(access_token)
            authorizer.set_refresh_cookies(refresh_token)

            return True
        except Exception as exc:
            print(exc)
            return False

    @staticmethod
    async def refresh_access_token(authorizer: AuthJWT) -> bool :
        authorizer.jwt_refresh_token_required()
        current_user = authorizer.get_jwt_subject()

        await JwtAuthUtils.revoke_token(authorizer)

        access_token_expires = timedelta(seconds=utils_settings.ACCESS_TOKEN_EXPIRE_SECONDS)
        add_claims = {'ip': '127.0.0.1'}  # TODO populate with user roles

        access_token = authorizer.create_access_token(subject=current_user, expires_time=access_token_expires,
            user_claims=add_claims)

        # Set the JWT and CSRF double submit cookies in the response
        try:
            authorizer.set_access_cookies(access_token)
            return True
        except Exception as exc:
            print(exc)
            raise

    @staticmethod
    async def login_success(user_json: dict, authorizer: AuthJWT):
        user_id = Utils.remove_dash(user_json['id'])
        user_json['id'] = user_id
        active_auditor: ActiveAuditor = await ActiveAuditorService.login(user_id, user_json)
        JwtAuthUtils.set_access_token(active_auditor, authorizer)
        user_json["bvn"] = None
        return LoginSuccessDto(**user_json)

    @staticmethod
    async def access_token_protected(token: str = Depends(oauth2_scheme), authorizer: AuthJWT = Depends()):
        authorizer.jwt_required()
        return authorizer

    @staticmethod
    def generate_pkce() -> tuple[str, str, str]:
        code_verifier = secrets.token_urlsafe(64)
        digest = hashlib.sha256(code_verifier.encode('utf-8')).digest()
        code_challenge = base64.urlsafe_b64encode(digest).rstrip(b'=').decode('utf-8')

        state  = secrets.token_urlsafe(16)

        return code_challenge, code_verifier, state



from datetime import timedelta, date
from logging import Logger
from typing import Optional, List

from appodus_utils import Page
from appodus_utils import Utils
from fastapi.encoders import jsonable_encoder
from kink import inject, di

from appodus_utils.common.auth_utils import JwtAuthUtils
from appodus_utils.db.models import SuccessResponse
from appodus_utils.db.redis_utils import RedisUtils
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.method_trace_logger import method_trace_logger
from appodus_utils.decorators.transactional import transactional
from appodus_utils.domain.user.auth.active_auditor.service import ActiveAuditorService
from appodus_utils.domain.user.auth.models import LoginSuccessDto, EmailValidationRequest
from appodus_utils.domain.user.models import CreateUserDto, _CreateUserDto, User, QueryUserDto, SearchUserDto, \
     _UpdateUserDto, UpdateNameDto, UserStatus, Gender, KYCAgent
from appodus_utils.domain.user.repo import UserRepo
from appodus_utils.domain.user.validator import UserValidator
from appodus_utils.exception.exceptions import InvalidTokenException
from appodus_utils.integrations.messaging.models import MessageRecipientUserId, MessageContextModule, \
    MessageRequestRecipient

logger: Logger = di['logger']

@inject
@decorate_all_methods(transactional(), exclude=['__init__', 'get_account_security_messages'], exclude_startswith='_')
@decorate_all_methods(method_trace_logger, exclude=['__init__', 'get_account_security_messages'], exclude_startswith='_')
class UserService:
    def __init__(self, user_repo: UserRepo,
                 user_validator: UserValidator,
                 active_auditor_service: ActiveAuditorService
                 ):
        self._user_repo = user_repo
        self._user_validator = user_validator
        self._active_auditor_service = active_auditor_service
        self._otp_token_expire_seconds = int(Utils.get_from_env_fail_if_not_exists("EMAIL_OTP_TOKEN_EXPIRE_SECONDS"))
        self._email_otp_token_expire_seconds = int(Utils.get_from_env_fail_if_not_exists("EMAIL_OTP_TOKEN_EXPIRE_SECONDS"))

    async def get_account_security_messages(self):
        """
        Returns an instance of AccountSecurityMessages from our dependency injector, kink.
        This method is here to avoid static dependency issues.
        :return:
        """
        from appodus_utils.domain.user.user_messages import AccountSecurityMessages

        return di[AccountSecurityMessages]


    async def create_user(self, obj_in: CreateUserDto) -> LoginSuccessDto:
        await self._user_validator.should_not_exist_by_email(obj_in.email)
        await self._validate_email_verification_otp(email=str(obj_in.email), otp=obj_in.otp)

        obj_in_copy = obj_in.model_copy()
        obj_in_copy.password = Utils.get_password_hash(obj_in.password)

        create_user_dto = _CreateUserDto(**obj_in_copy.model_dump())
        user_response = await self._user_repo.create(create_user_dto.model_dump(exclude={'otp', 'timezone', 'locale'}))
        user = user_response.data

        await self.post_create_user(created_user=user)

        user_data = await self.combine_profile_data(user)

        await (await self.get_account_security_messages()).send_new_user_welcome_message(
            recipient_user_id=MessageRecipientUserId(user_id=user.id),
            context_modules=[MessageContextModule.USER]
        )

        authorizer = await self._active_auditor_service.get_authorizer_from_context()
        return await JwtAuthUtils.login_success(user_data, authorizer)

    async def post_create_user(self, created_user: QueryUserDto):
        """
        Use to handle all post user creation operations like, settings and profile creation. Please override

        :param created_user:
        :return:
        """
        pass

    async def combine_profile_data(self, user: QueryUserDto):
        """
        Use to combine user and profile data for return user details after successful auth

        :param user:
        :return:
        """
        user_data = jsonable_encoder(user)
        return user_data

    async def get_user(self, user_id: str) -> SuccessResponse[QueryUserDto]:
        await self._user_validator.should_exist_by_id(user_id)
        return await self._user_repo.get(user_id)

    async def get_user_model(self, user_id: str, query_fields: Optional[str] = None) -> User:
        await self._user_validator.should_exist_by_id(user_id)
        return await self._user_repo.get_model(user_id, query_fields)

    async def exists_by_email(self, email: str) -> bool:
        return await self._user_repo.exists_by_email(email)

    async def get_by_email(self, email: str) -> Optional[QueryUserDto]:
        user = await self._user_repo.get_by_email(email)

        return user

    async def get_user_page(self, search_dto: SearchUserDto) -> Page[QueryUserDto]:
        return await self._user_repo.get_page(search_dto=search_dto)

    # async def update_by_admin(self, user_id: str, obj_in: Union[UpdateUserDto, Dict[str, Any]]) -> bool:
    #     await self._user_validator.should_exist_by_id(user_id)
    #     await self._user_repo.update(user_id, obj_in)
    #
    #     return True

    async def update_user_change_password(self, new_password: str) -> SuccessResponse[QueryUserDto]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()

        user_id = authorizer.get_jwt_subject()

        password_last_updated = Utils.datetime_now_to_db()
        obj_in = _UpdateUserDto(password=Utils.get_password_hash(new_password),
                                password_last_updated=password_last_updated)

        updated_user = await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        await (await self.get_account_security_messages()).send_password_updated_message(
            recipient_user_id=MessageRecipientUserId(user_id=user_id),
            context_modules=[MessageContextModule.USER])

        return updated_user

    async def update_last_active(self, user_id: str) -> SuccessResponse[QueryUserDto]:
        obj_in = _UpdateUserDto(last_active_date=Utils.datetime_now_to_db())
        return await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

    async def update_active_user_name(self, update_dto: UpdateNameDto) -> SuccessResponse[QueryUserDto]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()

        user_id = authorizer.get_jwt_subject()
        dto_data = jsonable_encoder(update_dto)
        obj_in = _UpdateUserDto(**dto_data)
        updated_user = await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        await (await self.get_account_security_messages()).send_name_updated_message(
            recipient_user_id=MessageRecipientUserId(user_id=user_id),
            context_modules=[MessageContextModule.USER])

        return updated_user

    async def soft_delete_by_admin(self, user_id: str) -> SuccessResponse[bool]:
        await self._user_validator.should_exist_by_id(user_id)
        await self._user_repo.soft_delete(user_id)

        return SuccessResponse(
            data=True
        )

    async def deactivate_active_user(self, reason: str) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()
        user_id = authorizer.get_jwt_subject()

        obj_in = _UpdateUserDto(status=UserStatus.DEACTIVATED, notes=reason)
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        await (await self.get_account_security_messages()).send_account_deactivation_message(
            recipient_user_id=MessageRecipientUserId(user_id=user_id),
            context_modules=[MessageContextModule.USER],
            extra_context={
                "reason": reason
            }
        )

        return SuccessResponse(
            data=True
        )

    async def activate_user(self, user_id: str, reason: str) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.jwt_required()

        obj_in = _UpdateUserDto(status=UserStatus.ACTIVE, notes=reason)
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        await (await self.get_account_security_messages()).send_account_activation_message(
            recipient_user_id=MessageRecipientUserId(user_id=user_id),
            context_modules=[MessageContextModule.USER],
            extra_context={
                "reason": reason
            }
        )

        return SuccessResponse(
            data=True
        )

    async def send_phone_validation_message(self, phone_ext: str, phone: str) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.jwt_required()
        user_id = authorizer.get_jwt_subject()
        otp = Utils.get_otp_code('i')

        time_to_live = timedelta(seconds=self._otp_token_expire_seconds)
        phone_delimited = f'{phone_ext}:{phone}'

        store_key = f'{user_id}:{otp}'
        await RedisUtils.set_redis(store_key, phone_delimited, time_to_live)


        await (await self.get_account_security_messages()).send_phone_verification_message(
            recipient_user_id=MessageRecipientUserId(user_id=user_id),
            context_modules=[MessageContextModule.USER],
            extra_context={
                "otp": otp,
                "phone_ext": phone_ext,
                "phone": phone
            }
        )

        return SuccessResponse(
            data=True
        )

    async def update_phone(self, otp: str) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()
        user_id = authorizer.get_jwt_subject()

        store_key = f'{user_id}:{otp}'
        phone_str = await RedisUtils.get_redis(store_key)
        if not phone_str:
            raise InvalidTokenException()

        phone_arr = phone_str.split(":")
        phone_ext = phone_arr[0]
        phone = phone_arr[1]

        obj_in = _UpdateUserDto(
            phone_ext=phone_ext,
            phone=phone,
            phone_validated=True
        )
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        return SuccessResponse(
            data=True
        )

    async def send_email_validation_message(self, payload: EmailValidationRequest, is_a_new_user: bool = True) -> SuccessResponse[bool]:
        if is_a_new_user:
            await self._user_validator.should_not_exist_by_email(payload.email)

        otp = Utils.get_otp_code()
        store_key = f'{payload.email}:{otp}'

        time_to_live = timedelta(seconds=self._otp_token_expire_seconds)
        await RedisUtils.set_redis(store_key, payload.email, time_to_live)

        if is_a_new_user:
            await (await self.get_account_security_messages()).send_direct_email_verification_message(
                recipient=MessageRequestRecipient(
                    user_id="pending_user_id",
                    fullname="Guest",
                    email=payload.email
                ),
                context={
                    "otp": otp,
                }
            )
        else:
            _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
            authorizer.fresh_jwt_required()
            user_id = authorizer.get_jwt_subject()

            await (await self.get_account_security_messages()).send_email_verification_message(
                recipient_user_id=MessageRecipientUserId(user_id=user_id),
                context_modules=[MessageContextModule.USER],
                extra_context={
                    "otp": otp,
                }
            )

        return SuccessResponse(
            data=True
        )

    @staticmethod
    async def _validate_email_verification_otp(email: str, otp: str) -> SuccessResponse[bool]:
        store_key = f'{email}:{otp}'
        new_email = await RedisUtils.get_redis(store_key)
        if not new_email:
            raise InvalidTokenException()

        return SuccessResponse(
            data=True
        )

    async def store_verified_email_otp(self, email: str, otp: str):
        store_key = f'{email}:{otp}'
        time_to_live = timedelta(minutes=self._email_otp_token_expire_seconds)
        await RedisUtils.set_redis(store_key, email, time_to_live)

    async def validate_and_store_email_verification_otp(self, email: str, otp: str) -> SuccessResponse[bool]:
        await self._validate_email_verification_otp(email=email, otp=otp)

        await self.store_verified_email_otp(email=email, otp=otp)

        return SuccessResponse(
            data=True
        )

    async def update_email(self, email: str, otp: str) -> SuccessResponse[bool]:
        await self._validate_email_verification_otp(email=email, otp=otp)

        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()
        user_id = authorizer.get_jwt_subject()

        obj_in = _UpdateUserDto(
            email=email,
            email_validated=True
        )
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        return SuccessResponse(
            data=True
        )

    async def update_gender(self, gender: Gender) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()
        user_id = authorizer.get_jwt_subject()

        obj_in = _UpdateUserDto(gender=gender)
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        return SuccessResponse(
            data=True
        )

    async def update_dob(self, dob: date) -> SuccessResponse[bool]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.fresh_jwt_required()
        user_id = authorizer.get_jwt_subject()


        obj_in = _UpdateUserDto(dob=dob)
        await self._user_repo.update(user_id, obj_in.model_dump(exclude_none=True))

        return SuccessResponse(
            data=True
        )

    async def get_logged_user_page(self, search_dto: SearchUserDto) -> Page[QueryUserDto]:
        return await self._user_repo.get_page(search_dto=search_dto)

    async def get_user_pending_kyc(self) -> List[KYCAgent]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        authorizer.jwt_required()
        user_id = authorizer.get_jwt_subject()

        user_response = await self.get_user(user_id=user_id)
        user = user_response.data

        pending_kyc = []

        if not user.email_validated:
            pending_kyc.append(KYCAgent.EMAIL)
        if not user.phone_validated:
            pending_kyc.append(KYCAgent.PHONE)
        if not user.bvn_validated:
            pending_kyc.append(KYCAgent.BVN)
        if not user.address_validated:
            pending_kyc.append(KYCAgent.ADDRESS)
        if not user.has_profile_picture:
            pending_kyc.append(KYCAgent.PROFILE_PICTURE)
        if not user.identity_validated:
            pending_kyc.append(KYCAgent.IDENTITY)

        return pending_kyc




    # async def update_active_user_picture(self, profile_picture: UploadFile) -> bool:
    #     profile_id = await self._active_auditor_service.get_profile_id()
    #
    #     file_dto = CreateStoredFileDto(
    #         owner=FileOwner.GUEST,
    #         owner_id=user_id,
    #         type=FileType.PROFILE,
    #         files=[profile_picture]
    #     )
    #
    #     if background_tasks:
    #         background_tasks.add_task(self._stored_file_service.create_file, file_dto, authorizer, background_tasks)
    #     else:
    #         await self._stored_file_service.create_file(file_dto, authorizer, background_tasks)
    #
    #     await self._profile_validator.should_exist_by_id(profile_id)
    #     obj_in = _UpdateProfileDto(has_profile_picture=True)
    #     updated = await self._profile_repo.update(profile_id, obj_in.model_dump(exclude_none=True))
    #
    #     return updated.has_profile_picture
    #
    # async def update_active_user_selfie(self, profile_selfie: UploadFile,
    #                                     background_tasks: BackgroundTasks = None) -> bool:
    #     profile_id = await self._active_auditor_service.get_profile_id()
    #     await self._profile_validator.should_exist_by_id(profile_id)
    #
    #     file_dto = CreateStoredFileDto(
    #         owner=FileOwner.GUEST,
    #         owner_id=user_id,
    #         type=FileType.SELFIE,
    #         files=[profile_selfie]
    #     )
    #
    #     if background_tasks:
    #         background_tasks.add_task(self._stored_file_service.create_file, file_dto, authorizer, background_tasks)
    #     else:
    #         await self._stored_file_service.create_file(file_dto, authorizer, background_tasks)
    #
    #     obj_in = _UpdateProfileDto(has_selfie_picture=True)
    #     updated = await self._profile_repo.update(profile_id, obj_in.model_dump(exclude_none=True))
    #
    #     return updated.has_selfie_picture



    # async def update_active_user_bvn(self, bvn: str) -> bool:
    #     profile_id = await self._active_auditor_service.get_profile_id()
    #     await self._profile_validator.should_exist_by_id(profile_id)
    #     obj_in = _UpdateProfileDto(bvn=bvn)
    #     updated = await self._profile_repo.update(profile_id, obj_in.model_dump(exclude_none=True))
    #     return updated.bvn == bvn
    #
    # async def bvn_validated(self) -> bool:
    #     profile_id = await self._active_auditor_service.get_profile_id()
    #     await self._profile_validator.should_exist_by_id(profile_id)
    #     obj_in = _UpdateProfileDto(bvn_validated=True)
    #     updated = await self._profile_repo.update(profile_id, obj_in.model_dump(exclude_none=True))
    #     return updated.bvn_validated
    #
    # async def identity_validated(self) -> bool:
    #     profile_id = await self._active_auditor_service.get_profile_id()
    #     await self._profile_validator.should_exist_by_id(profile_id)
    #     obj_in = _UpdateProfileDto(identity_validated=True)
    #     updated = await self._profile_repo.update(profile_id, obj_in.model_dump(exclude_none=True))
    #
    #     return updated.identity_validated



from abc import abstractmethod, ABC
from typing import Dict, Any

from appodus_utils.config.settings import TemplatingEngine


class ITemplateEngineProvider(ABC):
    """Abstract base class for all template engines"""

    @property
    @abstractmethod
    def template_engine(self) -> TemplatingEngine:
        pass

    @property
    @abstractmethod
    def template_extension(self) -> str:
        pass

    @abstractmethod
    def render(self, template_name: str, context: Dict[str, Any]) -> str:
        """Render a template with the given context"""
        pass

    @abstractmethod
    def supports_template(self, template_name: str) -> bool:
        """Check if template exists in the engine"""
        pass



from http import HTTPStatus

from fastapi import APIRouter
from kink import di

from appodus_utils.db.models import SuccessResponse, Page
from appodus_utils.domain.user.address.models import QueryAddressDto, CreateAddressDto, SearchAddressDto, \
    UpdateAddressDto
from appodus_utils.domain.user.address.service import AddressService

address_service: AddressService = di[AddressService]

address_router = APIRouter(prefix="/addresses", tags=["Addresses"])


@address_router.post('/', response_model=SuccessResponse[QueryAddressDto], status_code=HTTPStatus.CREATED)
async def create_address(create_dto: CreateAddressDto) -> SuccessResponse[QueryAddressDto]:
    return await address_service.create_address(create_dto=create_dto)


@address_router.get('/{address_id}', response_model=SuccessResponse[QueryAddressDto], status_code=HTTPStatus.OK)
async def get_address(address_id: str) -> SuccessResponse[QueryAddressDto]:
    return await address_service.get_address(address_id=address_id)


@address_router.get('/', response_model=Page[QueryAddressDto], status_code=HTTPStatus.OK)
async def get_address_page(search_dto: SearchAddressDto) -> Page[QueryAddressDto]:
    return await address_service.get_address_page(search_dto=search_dto)


@address_router.get('/active-user', response_model=QueryAddressDto, status_code=HTTPStatus.OK)
async def get_address_page_for_active_user():
    return await address_service.get_address_page_for_active_user()


@address_router.patch('/{address_id}', response_model=SuccessResponse[QueryAddressDto], status_code=HTTPStatus.OK)
async def update_address(address_id: str, update_dto: UpdateAddressDto) -> SuccessResponse[QueryAddressDto]:
    return await address_service.update_address(address_id=address_id, update_dto=update_dto)


@address_router.patch('/{address_id}/verify', response_model=SuccessResponse[QueryAddressDto], status_code=HTTPStatus.OK)
async def verify_address(address_id: str) -> SuccessResponse[QueryAddressDto]:
    return await address_service.verify_address(address_id=address_id)


@address_router.delete('/{address_id}', response_model=SuccessResponse[bool], status_code=HTTPStatus.OK)
async def soft_delete_address(address_id: str) -> SuccessResponse[bool]:
    return await address_service.soft_delete_address(address_id=address_id)



from typing import Type

from kink import inject
from sqlalchemy.ext.asyncio import AsyncSession

from appodus_utils.db.repo import GenericRepo
from appodus_utils.domain.user.address.models import Address, CreateAddressDto, UpdateAddressDto, QueryAddressDto, \
    SearchAddressDto


@inject
class AddressRepo(GenericRepo[Address, CreateAddressDto, UpdateAddressDto, QueryAddressDto, SearchAddressDto]):
    def __init__(self, db: AsyncSession, model: Type[Address] = Address, query_dto: Type[QueryAddressDto] = QueryAddressDto):
        super().__init__(db, model, query_dto)
        self.db = db



import enum
from typing import Optional

from httpx import AsyncClient, ReadTimeout
from kink import di
from starlette import status

from appodus_utils import Utils
from appodus_utils.exception.exceptions import InternalServerException

from appodus_utils.sdk.appodus_sdk.services.chatbot_client import ChatbotClient
from appodus_utils.sdk.appodus_sdk.services.contract_client import ContractClient
from appodus_utils.sdk.appodus_sdk.services.document_client import DocumentClient
from appodus_utils.sdk.appodus_sdk.services.message_client import MessageClient
from appodus_utils.sdk.appodus_sdk.utils import AppodusClientUtils


class ApiVersion(str, enum.Enum):
    V1 = "v1"


class AppodusClient:
    """
    This is a utility class that abstracts the various Appodus utility services:
    [document_manager, contract_manager, communication_manager]
    """

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        api_version: ApiVersion = ApiVersion.V1,
        service_url: Optional[str] = None,
        http_client: Optional[AsyncClient] = None,
    ) -> None:
        """
        :param client_id: Optional override for APPODUS_CLIENT_ID. OAuth client ID for authentication.
        :param client_secret: Optional override for APPODUS_CLIENT_SECRET. OAuth client secret
        :param api_version: API version to use (e.g., ApiVersion.V1)
        :param service_url: Optional override for APPODUS_SERVICES_URL
        :param http_client: Optional injected AsyncClient for testing
        """
        self._appodus_service_url = service_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._api_version = api_version
        self._http_client = http_client or di[AsyncClient]

        self.client_utils: Optional[AppodusClientUtils] = None

        self._contract_client: Optional[ContractClient] = None
        self._document_client: Optional[DocumentClient] = None
        self._chatbot_client: Optional[ChatbotClient] = None
        self._message_client: Optional[MessageClient] = None

    def init(self) -> 'AppodusClient':
        """Initializes Appodus service clients and checks connectivity."""
        try:
            self._load_config()
            self._create_client_utils()
            self._init_clients()
            # await self._check_service_health()
        except Exception as e:
            raise InternalServerException(message=f"Unable to initialize AppodusClient: {e.__str__()}")

        return self

    @property
    def messaging(self) -> MessageClient:
        return self._require_initialized(self._message_client, "messaging")

    @property
    def contract(self) -> ContractClient:
        return self._require_initialized(self._contract_client, "contract")

    @property
    def document(self) -> DocumentClient:
        return self._require_initialized(self._document_client, "document")

    @property
    def chatbot(self) -> ChatbotClient:
        return self._require_initialized(self._chatbot_client, "chatbot")

    def _load_config(self) -> None:
        if not self._appodus_service_url:
            self._appodus_service_url = Utils.get_from_env_fail_if_not_exists('APPODUS_SERVICES_URL')
        if not self._client_id:
            self._client_id = Utils.get_from_env_fail_if_not_exists('APPODUS_CLIENT_ID')
        if not self._client_secret:
            self._client_secret = Utils.get_from_env_fail_if_not_exists('APPODUS_CLIENT_SECRET')

    def _create_client_utils(self) -> None:
        self.client_utils = AppodusClientUtils(
            client_id=self._client_id,
            client_secret=self._client_secret,
            api_version=self._api_version.value,
            http_client=self._http_client
        )

    def _init_clients(self) -> None:
        self._message_client = MessageClient(
            message_manager_url=self._appodus_service_url,
            client_utils=self.client_utils
        )
        self._contract_client = ContractClient(
            contract_manager_url=self._appodus_service_url,
            client_utils=self.client_utils
        )
        self._document_client = DocumentClient(
            document_manager_url=self._appodus_service_url,
            client_utils=self.client_utils
        )
        self._chatbot_client = ChatbotClient(
            chatbot_manager_url=self._appodus_service_url,
            client_utils=self.client_utils
        )

    async def _check_service_health(self) -> None:
        error_message = f"The provided APPODUS_SERVICES_URL '{self._appodus_service_url}' is not reachable"
        try:
            response = await self._http_client.get(f"{self._appodus_service_url}/health")
            if response.status_code != status.HTTP_200_OK:
                raise InternalServerException(
                    message=f"The provided APPODUS_SERVICES_URL '{self._appodus_service_url}' is not reachable."
                )

            response.raise_for_status()
        except ReadTimeout as re:
            error_message = f"{error_message}: Request timeout"
            raise InternalServerException(
                message=error_message
            )
        except Exception as e:
            error_message = f"{error_message}: {e.__str__()}"
            raise InternalServerException(
                message=error_message
            )


    @staticmethod
    def _require_initialized(client, name: str):
        if client is None:
            raise InternalServerException(
                f"AppodusClient.{name} not initialized. Call `await client.init()` first."
            )
        return client



from logging import Logger

from appodus_utils import Page
from kink import inject, di

from appodus_utils.db.models import SuccessResponse
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.method_trace_logger import method_trace_logger
from appodus_utils.decorators.transactional import transactional
from appodus_utils.domain.user.device.models import CreateDeviceDto, QueryDeviceDto, SearchDeviceDto, _UpdateDeviceDto, \
    _CreateDeviceDto
from appodus_utils.domain.user.device.repo import DeviceRepo
from appodus_utils.domain.user.device.validator import DeviceValidator
from appodus_utils.integrations.messaging.models import PushToken

logger: Logger = di['logger']


@inject
@decorate_all_methods(transactional(), exclude=['__init__'], exclude_startswith='_')
@decorate_all_methods(method_trace_logger, exclude=['__init__'], exclude_startswith='_')
class DeviceService:
    def __init__(self, device_repo: DeviceRepo, device_validator: DeviceValidator):
        self._device_repo = device_repo
        self._device_validator = device_validator

    async def create_device(self, obj_in: CreateDeviceDto) -> SuccessResponse[QueryDeviceDto]:
        create_dto = _CreateDeviceDto.model_validate(obj_in.model_dump())
        return await self._device_repo.create(create_dto)

    async def get_device_page(self, search_dto: SearchDeviceDto) -> Page[QueryDeviceDto]:
        return await self._device_repo.get_page(search_dto=search_dto)

    async def update_device_push_token(self, device_id: str, push_token: PushToken) -> SuccessResponse[bool]:
        await self._device_validator.should_exist_by_id(device_id)

        obj_in = _UpdateDeviceDto(push_token=push_token)
        await self._device_repo.update(device_id, obj_in.model_dump(exclude_none=True))

        return SuccessResponse(
            data=True
        )



from unittest import IsolatedAsyncioTestCase

import respx
from _pytest.monkeypatch import MonkeyPatch
from fastapi.encoders import jsonable_encoder
from kink import di

from appodus_utils import Utils

from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.domain.contract.models import GenerateAndSendSignDto
from appodus_utils.integrations.document_sign.models import Signer, SignActionType, DocumentVerificationType, \
    SignRequestResponseDto
from appodus_utils.sdk.appodus_sdk.appodus import AppodusClient
from appodus_utils.sdk.appodus_sdk.services.contract_client import ContractClient
from tests.appodus_utils.test_utils import mock_http_request


class TestAppodusContractClient(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.appodus_client: AppodusClient = di[AppodusClient]
        self.monkeypatch = MonkeyPatch()
        appodus_service_url = Utils.get_from_env_fail_if_not_exists('APPODUS_SERVICES_URL')
        self.url_endpoint = f"{appodus_service_url}/v1/contracts"
        self.health_check_endpoint = f"{appodus_service_url}/health"


    async def asyncTearDown(self):
        self.monkeypatch.undo()
        # await self.appodus_client.client_utils.get_http_client.aclose()

    @respx.mock(assert_all_called=False, assert_all_mocked=False)
    async def test_generate_and_send_sign_request(self):
        # PREPARE
        contract_client: ContractClient = self.appodus_client.contract

        payload = GenerateAndSendSignDto(
            template_id="template_id",
            template_variables={
                "Agreement_Date": Utils.datetime_now_format(output_format="%d/%m/%Y"),
                "Appodus_Address": "23 Agodogba Street, Parkview, Ikoyi, Lagos NG.",
                "Client_Company_Name": "Arone Technologies Limited",
                "Client_Address": "4 Mbanefo Street, New Haven, Enugu",
                "Client_Phone": "2348056453738",
                "Client_Email": "kingsley.ezenwere@gmail.com"
            },
            contract_id="contract_id",
            signers=[
                Signer(
                    action_type=SignActionType.SIGN,
                    recipient_email="kingsley.ezenwere@gmail.com",
                    recipient_phonenumber="+2347039018727",
                    recipient_name="Kingsley Ezenwere",
                    verify_recipient=True,
                    verification_type=DocumentVerificationType.EMAIL,
                    signing_order=0
                )
            ],
            signing_client_name="Gabrielle"
        )
        requests_json = jsonable_encoder(payload)
        self.url_endpoint = f"{self.url_endpoint}"
        request_headers = self.appodus_client.client_utils.auth_headers(
            "post",
            self.url_endpoint,
            requests_json
        )

        return_json = SignRequestResponseDto(
            request_id="request_id",
            signers=payload.signers
        )
        mock_http_request(
            monkeypatch=self.monkeypatch,
            http_client=self.appodus_client.client_utils.get_http_client,
            request_url=self.url_endpoint,
            http_method="post",
            request_headers=request_headers,
            return_json=return_json.model_dump(),
            _json=requests_json
        )

        client_response = await contract_client.generate_and_send_sign_request(request_dto=payload)

        self.assertEqual(return_json, client_response)

    @respx.mock(assert_all_called=False, assert_all_mocked=False)
    async def test_get_signer_embedded_sign_url(self):
        # PREPARE
        contract_client: ContractClient = self.appodus_client.contract
        signer_email = "kingsley.ezenwere@gmail.com"
        request_headers = self.appodus_client.client_utils.auth_headers(
            "get",
            f"{self.url_endpoint}"
        )
        self.url_endpoint = f"{self.url_endpoint}/signers/{signer_email}/embedded"

        mock_http_request(
            monkeypatch=self.monkeypatch,
            http_client=self.appodus_client.client_utils.get_http_client,
            request_url=self.url_endpoint,
            http_method="get",
            request_headers=request_headers
        )

        await contract_client.get_signer_embedded_sign_url(signer_email=signer_email)



import enum
from typing import Optional

from pydantic import Field

from appodus_utils import Object, Utils
from appodus_utils.domain.user.models import QueryUserDto


class SocialAuthOperationType(str, enum.Enum):
    LOGIN = "login"
    SIGNUP = "signup"

class EmailLoginRequest(Object):
    username: str
    password: str


class PasswordResetRequest(Object):
    email: str

class EmailValidationRequest(Object):
    email: str


class PasswordResetConfirm(Object):
    token: str
    new_password: str


class ChangePasswordDto(Object):
    old_password: str
    new_password: str


class SocialLoginUserInfo(Object):
    id: Optional[str] = None
    email: Optional[str] = None
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    exp: Optional[int] = 0
    otp: str = Field(default=Utils.random_str(8), description="Token used for email verification during user creation")
    operation_type: Optional[SocialAuthOperationType] = Field(None, description="Used to distinguish between a login/signup operation")


class OAuthCallbackRequest(Object):
    code: str
    code_verifier: str
    operation_type: SocialAuthOperationType


class Token(Object):
    access_token: str
    refresh_token: str
    token_type: str


class TokenType(str, enum.Enum):
    ACCESS = 'ACCESS'
    REFRESH = 'REFRESH'


class Role(str, enum.Enum):
    ADMIN = 'ADMIN'
    USER = 'USER'
    SELLER = 'SELLER'
    BUYER = 'BUYER'
    AGENT = 'AGENT'


class LoginSuccessDto(QueryUserDto):
    pass



from abc import ABC, abstractmethod

from starlette.requests import Request

from appodus_utils.domain.user.auth.models import SocialAuthOperationType, OAuthCallbackRequest, SocialLoginUserInfo
from appodus_utils.domain.user.auth.social_login.models import SocialAuthPlatform


class ISocialAuthProvider(ABC):
    @property
    @abstractmethod
    def platform(self) -> SocialAuthPlatform:
        pass

    @abstractmethod
    async def initialize(self, operation_type: SocialAuthOperationType, request: Request) -> dict:
        pass

    @abstractmethod
    async def verify(self, payload: OAuthCallbackRequest, request: Request) -> SocialLoginUserInfo:
        pass



from http import HTTPStatus
from typing import Dict, Any

from appodus_utils import Page
from fastapi import APIRouter
from kink import di

from appodus_utils.db.models import SuccessResponse
from appodus_utils.domain.user.device.models import CreateDeviceDto, QueryDeviceDto, SearchDeviceDto
from appodus_utils.domain.user.device.service import DeviceService
from appodus_utils.integrations.messaging.models import PushToken

device_service: DeviceService = di[DeviceService]
device_router = APIRouter(prefix="/devices", tags=["Devices"])


@device_router.post('/', response_model=SuccessResponse[QueryDeviceDto], status_code=HTTPStatus.CREATED)
async def create_device(create_dto: CreateDeviceDto) -> SuccessResponse[QueryDeviceDto]:
    return await device_service.create_device(create_dto)


@device_router.get('/', response_model=Page[QueryDeviceDto], status_code=HTTPStatus.OK)
async def get_device_page(search_dto: SearchDeviceDto) -> Page[QueryDeviceDto]:
    return await device_service.get_device_page(search_dto)


@device_router.put('/{device_id}', response_model=SuccessResponse[bool], status_code=HTTPStatus.OK)
async def update_device_push_token(device_id: str, push_token: PushToken) -> SuccessResponse[bool]:
    return await device_service.update_device_push_token(device_id, push_token)



from typing import Optional


from sqlalchemy import Column, String, Boolean

from appodus_utils import BaseEntity, PageRequest, BaseQueryDto
from appodus_utils import Object


class Currency(BaseEntity):
    __tablename__ = 'currencies'
    name = Column(String, nullable=False)
    code = Column(String, nullable=False)
    country_code = Column(String, nullable=False)
    is_active = Column(Boolean, nullable=False)


class CurrencyBaseDto(Object):
    name: str
    code: str
    country_code: str


class CreateCurrencyDto(CurrencyBaseDto):
    is_active: bool = True


class UpdateCurrencyDto(CurrencyBaseDto):
    name: Optional[str]
    code: Optional[str]
    country_code: Optional[str]


class _UpdateCurrencyDto(Object):
    is_active: Optional[bool]


class SearchCurrencyDto(PageRequest, BaseQueryDto):
    name: Optional[str] = None
    code: Optional[str] = None
    country_code: Optional[str] = None
    is_active: Optional[bool] = None


class QueryCurrencyDto(CurrencyBaseDto, BaseQueryDto):
    is_active: Optional[bool]



from kink import inject

from appodus_utils.domain.user.address.repo import AddressRepo
from appodus_utils.exception.exceptions import ResourceNotFoundException


@inject
class AddressValidator:
    def __init__(self, address_repo: AddressRepo):
        self._address_repo = address_repo

    async def should_exist_by_id(self, _id: str):
        if not (await self._address_repo.exists_by_id(_id)):
            raise ResourceNotFoundException("Address")



from contextlib import asynccontextmanager
from logging import Logger

from libre_fastapi_jwt.exceptions import AuthJWTException
from starlette import status




from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap

from appodus_utils.db.session import close_db_engine
from appodus_utils.domain import utils_router
from appodus_utils.domain.client.controller import client_router
from appodus_utils.domain.user.auth.active_auditor.global_context import init_auth_context
from appodus_utils.exception.exception_handlers import (
    appodus_exception_handler,
    http_error_handler,
    validation_exception_handler,
    generic_exception_handler,
    authjwt_exception_handler
)
from appodus_utils.exception.exceptions import AppodusBaseException
from appodus_utils.middleware.db_session_middleware import DBSessionMiddleware
from appodus_utils.middleware.request_logging_middleware import RequestLoggingMiddleware
from fastapi import FastAPI, Depends
from fastapi.exceptions import RequestValidationError
from httpx import AsyncClient
from kink import di
from starlette.exceptions import HTTPException

logger: Logger = di['logger']
httpx_client: AsyncClient = di[AsyncClient]


@asynccontextmanager
async def lifespan_event(app: FastAPI):
    logger.debug("Running lifespan..")

    logger.debug("Done running lifespan")
    yield
    await close_db_engine()
    await httpx_client.aclose()


fast_api_app = FastAPI(lifespan=lifespan_event, dependencies=[Depends(init_auth_context)])

# Routers
fast_api_app.include_router(client_router)
fast_api_app.include_router(utils_router)

# Exception Handlers
# Custom appodus exceptions
fast_api_app.add_exception_handler(AppodusBaseException, appodus_exception_handler)
# AuthJWTException
fast_api_app.add_exception_handler(AuthJWTException, authjwt_exception_handler)
# FastAPI built-in ones
fast_api_app.add_exception_handler(HTTPException, http_error_handler)
fast_api_app.add_exception_handler(RequestValidationError, validation_exception_handler)
# Catch-all fallback
fast_api_app.add_exception_handler(Exception, generic_exception_handler)
#
# # Middlewares
# app.add_middleware(ClientAuthMiddleware)
fast_api_app.add_middleware(DBSessionMiddleware)
fast_api_app.add_middleware(RequestLoggingMiddleware)


@fast_api_app.get("/health", status_code=status.HTTP_200_OK)
def health_check():
    return {"status": "healthy"}


if __name__ == "__main__":
    import uvicorn
    # logger.logger.error("Starting dev server:")
    uvicorn.run(fast_api_app, host="127.0.0.1", port=8000)



import json
from typing import List

from fastapi.encoders import jsonable_encoder

from appodus_utils.integrations.messaging.models import MessageRequest, BatchResult
from appodus_utils.sdk.appodus_sdk.utils import AppodusClientUtils


class MessagingService:
    def __init__(self, message_manager_url: str, client_utils: AppodusClientUtils):
        self._client_utils = client_utils
        self._message_manager_url = message_manager_url

    async def send_message(self, message_request: MessageRequest) -> BatchResult:
        return await self.send_bulk([message_request])

    async def send_bulk(self, message_requests: List[MessageRequest]) -> BatchResult:
        endpoint = f"{self._message_manager_url}/{self._client_utils.get_api_version}/messages"
        message_requests_data = jsonable_encoder(message_requests)
        headers = self._client_utils.auth_headers("post", f"{endpoint}", message_requests_data)
        response = await self._client_utils.get_http_client.post(f"{endpoint}", json=message_requests_data, headers=headers)
        response.raise_for_status()

        print(f"response text: {response.text}")
        response = response.json()
        return BatchResult(**response)



from fastapi import Depends, BackgroundTasks
from kink import di
from libre_fastapi_jwt import AuthJWT, AuthJWTBearer
from starlette.requests import Request

from appodus_utils.common.context_utils import ContextUtils
from appodus_utils.domain.user.auth.active_auditor.service import ActiveAuditorService

auth_JWT_bearer: AuthJWTBearer = di[AuthJWTBearer]
active_auditor_service: ActiveAuditorService = di[ActiveAuditorService]

async def init_auth_context(request: Request, background_tasks: BackgroundTasks, authorizer: AuthJWT = Depends(auth_JWT_bearer)):
    di[ContextUtils]  = lambda _di: ContextUtils(request)

    background_tasks_context_token, authorizer_context_token, active_auditor_context_token = await active_auditor_service.init_context(
        background_tasks, authorizer)
    yield
    await active_auditor_service.reset_context(background_tasks_context_token, authorizer_context_token, active_auditor_context_token)



import urllib.parse
from logging import Logger
from urllib.parse import urlunparse, urljoin

from kink import di
from starlette.requests import Request

from appodus_utils import Utils
from appodus_utils.common.auth_utils import JwtAuthUtils
from appodus_utils.db.redis_utils import RedisUtils
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.method_trace_logger import method_trace_logger
from appodus_utils.domain.user.auth.models import SocialAuthOperationType
from appodus_utils.domain.user.auth.social_login.models import SocialAuthPlatform

logger: Logger = di["logger"]


@decorate_all_methods(method_trace_logger, exclude=['__init__'], exclude_startswith='_')
class OauthUtils:

    @staticmethod
    async def init_0auth(request: Request, base_url: str, client_id: str, redirect_uri: str, scope: str, operation_type: SocialAuthOperationType) -> dict:
        code_challenge, code_verifier, state = JwtAuthUtils.generate_pkce()

        params = {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "response_type": "code",
            "scope": scope,
            "state": state,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
        }

        await RedisUtils.set_redis(state, {code_verifier: code_verifier, operation_type: operation_type})

        query_string = urllib.parse.urlencode(params)
        return {
            "url": f"{base_url}?{query_string}",
            "code_verifier": code_verifier
        }

    @staticmethod
    async def get_auth_redirect_url(platform: SocialAuthPlatform, request: Request) -> str:
        """
        The redirect url is similar to the init url, when we remove the tailing /init
        :param platform:
        :param request:
        :return:
        """
        # Parse the incoming URL parts
        base_path = request.url.path.rsplit("/", 1)[0] + "/" # Remove the /init in '/v1/users/auths/socials/google/init'

        full_url = urlunparse((
            request.url.scheme,
            request.url.netloc,
            base_path,
            "",  # params
            "",  # query
            ""  # fragment
        ))

        return full_url



from unittest import IsolatedAsyncioTestCase

import respx
from _pytest.monkeypatch import MonkeyPatch
from kink import di

from appodus_utils.integrations.messaging.models import MessageRequest, MessageChannel, MessageRequestRecipient, \
    EmailPayload, EmailParty, BatchResult
from fastapi.encoders import jsonable_encoder

from appodus_utils import Utils
from appodus_utils.sdk.appodus_sdk.appodus import AppodusClient
from appodus_utils.sdk.appodus_sdk.services.message_client import MessageClient
from tests.appodus_utils.test_utils import mock_http_request


class TestAppodusMessageClient(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.appodus_client: AppodusClient = di[AppodusClient]
        self.monkeypatch = MonkeyPatch()
        appodus_service_url = Utils.get_from_env_fail_if_not_exists('APPODUS_SERVICES_URL')
        self.url_endpoint = f"{appodus_service_url}/v1/messages"
        self.health_check_endpoint = f"{appodus_service_url}/health"


    async def asyncTearDown(self):
        self.monkeypatch.undo()

    @respx.mock(assert_all_called=False, assert_all_mocked=False)
    async def test_send_bulk(self):
        # PREPARE
        message_client: MessageClient = self.appodus_client.messaging

        message_request = (MessageRequest.builder()
                           .channel(MessageChannel.EMAIL)
                           .to(MessageRequestRecipient(
            user_id="user_id",
            fullname="Kingsley Ezenwere",
            email="kingsley.ezenwere@gmail.com"
        ))
                           .payload(
            EmailPayload(
                html="<strong>Appodus Email Test drive</strong>",
                subject="Hello from appodus",
                email_from=EmailParty(email="festus.ezenwere@gmail.com", fullname="Festus Ezenwere"),
                reply_to=EmailParty(email="appodus@gmail.com", fullname="appodus global")
            )).build())

        payload = [message_request]

        requests_json = jsonable_encoder(payload)
        self.url_endpoint = f"{self.url_endpoint}"
        request_headers = self.appodus_client.client_utils.auth_headers(
            "post",
            self.url_endpoint,
            requests_json
        )

        return_json = BatchResult(
            total=1,
            successes=["pass"],
            failures=["fail"],
            processing_time=323432
        )

        mock_http_request(
            monkeypatch=self.monkeypatch,
            http_client=self.appodus_client.client_utils.get_http_client,
            request_url=self.url_endpoint,
            http_method="post",
            request_headers=request_headers,
            return_json=return_json.model_dump(),
            _json=requests_json
        )

        client_response = await message_client.send_bulk(message_requests=payload)

        self.assertEqual(return_json, client_response)



from fastapi import APIRouter



from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap

from appodus_utils import RouterUtils
from appodus_utils.domain.key_value.models import KeyValue

utils_router = APIRouter(prefix="/v1", tags=["Utils"])

RouterUtils.add_routers(utils_router, [])


from typing import Type, List

from httpx import AsyncClient
from kink import di
from libre_fastapi_jwt import AuthJWTBearer
from redis.asyncio import Redis

from appodus_utils.config.logger import LoggerFactory

bootstrap_di_called = False


class BaseDiBootstrap:

    def inject_redis(self):
        """ PLease override in the inheriting class """
        di[Redis] = lambda _di: {}

    def inject_others(self):
        """ PLease implement in the inheriting class """
        pass

    def init(self) -> None:
        self.inject_redis()
        self.inject_others()

        di['logger'] = lambda _di: LoggerFactory().get_logger()
        di[AuthJWTBearer] = lambda _di: AuthJWTBearer()
        di[AsyncClient] = lambda _di: AsyncClient(timeout=30.0, follow_redirects=True)

        from appodus_utils.sdk.appodus_sdk.appodus import AppodusClient
        di[AppodusClient] = AppodusClient().init()

    @staticmethod
    def register_all_subclasses(base_cls: Type):
        instances = BaseDiBootstrap._get_all_subclasses_instances(base_cls)

        di[List[base_cls]] = instances
        return instances


    @staticmethod
    def register_subclass_if_exists(base_cls: Type):
        instances = BaseDiBootstrap._get_all_subclasses_instances(base_cls)

        if not instances:
            instance = di[base_cls]  # Use DI container to resolve dependencies
            instances.append(instance)

        di[base_cls] = instances[-1]
        return instances

    @staticmethod
    def _get_all_subclasses_instances(base_cls: Type):
        subclasses: List = base_cls.__subclasses__()
        instances = []
        for subclass in subclasses:
            other_subclasses = subclass.__subclasses__()
            if len(other_subclasses) > 0:
                _instances = BaseDiBootstrap._get_all_subclasses_instances(subclass)
                instances.extend(_instances)
            else:
                instance = di[subclass]  # Use DI container to resolve dependencies
                instances.append(instance)

        return instances


if not bootstrap_di_called:
    base_di_bootstrap = BaseDiBootstrap()
    base_di_bootstrap.init()

    bootstrap_di_called = True



from unittest import IsolatedAsyncioTestCase

from _pytest.monkeypatch import MonkeyPatch
from kink import di
from polyfactory.factories.pydantic_factory import ModelFactory


from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap

from appodus_utils.common.appodus_test_utils import TestUtils
from appodus_utils.db.session import close_db_engine
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.transactional import TransactionSessionPolicy, transactional
from appodus_utils.domain.user.address.models import Address, CreateAddressDto, QueryAddressDto, SearchAddressDto, \
    UpdateAddressDto
from appodus_utils.domain.user.address.service import AddressService
from appodus_utils.domain.user.auth.models import LoginSuccessDto
from appodus_utils.domain.user.models import CreateUserDto, QueryUserDto, User
from appodus_utils.domain.user.service import UserService
from tests.appodus_utils.domain.user.test_e2e_service import CreateUserDtoFactory
from tests.appodus_utils.test_utils import mock_active_auditor_service


@decorate_all_methods(transactional(session_policy=TransactionSessionPolicy.ALWAYS_NEW),
                      exclude=['__init__', 'asyncTearDown', 'asyncSetUp'], exclude_startswith='_')
class TestAddressService(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.address_service: AddressService = di[AddressService]
        self.user_service: UserService = di[UserService]
        self.monkeypatch = MonkeyPatch()

    async def asyncTearDown(self):
        self.monkeypatch.undo()
        await self._truncate_tables()
        await close_db_engine()

    @staticmethod
    async def _truncate_tables():
        await TestUtils.truncate_entities([Address, User])

    async def create_user_and_mock_active_auditor_service(self) -> LoginSuccessDto:

        create_user_dto = CreateUserDtoFactory.build()
        # mock_active_auditor_service
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(create_user_dto))

        await self.user_service.store_verified_email_otp(
            email=create_user_dto.email,
            otp=create_user_dto.otp
        )

        created_user_response = await self.user_service.create_user(create_user_dto)

        # re-mock_active_auditor_service to include User ID
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(created_user_response))

        return created_user_response

    async def create_address(self, user_id: str) -> QueryAddressDto:
        create_address_dto = CreateAddressDtoFactory.build()
        create_address_dto.user_id = user_id

        created_address_response = await self.address_service.create_address(create_address_dto)
        created_address = created_address_response.data

        return created_address


    async def test_create_address(self):
        create_address_dto = CreateAddressDtoFactory.build()

        created_address_response = await self.address_service.create_address(create_address_dto)
        created_address = created_address_response.data

        self.assertIsNotNone(created_address.id)

    async def test_get_address(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        created_address = await self.create_address(user_id=created_user.id)

        get_address_response = await self.address_service.get_address(address_id=created_address.id)
        gotten_address = get_address_response.data

        self.assertIsNotNone(gotten_address.id)

    async def test_get_address_page(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        count = 2
        for _ in range(count):
            await self.create_address(user_id=created_user.id)

        search_dto: SearchAddressDto = SearchAddressDto(
            page_size=10,
            page=0
        )

        address_page = await self.address_service.get_address_page(search_dto=search_dto)

        self.assertEqual(count, address_page.meta.count)

    async def test_get_address_page_for_active_user(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        count = 2
        for _ in range(count):
            await self.create_address(user_id=created_user.id)

        address_page = await self.address_service.get_address_page_for_active_user()

        self.assertEqual(count, address_page.meta.count)

    async def test_update_address(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        created_address = await self.create_address(user_id=created_user.id)

        update_dto = UpdateAddressDto(**created_address.model_dump())
        update_dto.lga = "Eti Osa"

        update_address_response = await self.address_service.update_address(
            address_id=created_address.id,
            update_dto=update_dto
        )
        updated_address = update_address_response.data

        self.assertEqual(update_dto.lga, updated_address.lga)

    async def test_verify_address(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        created_address = await self.create_address(user_id=created_user.id)

        get_address_response = await self.address_service.verify_address(address_id=created_address.id)
        gotten_address = get_address_response.data

        self.assertTrue(gotten_address.verified)

    async def test_soft_delete_address(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        created_address = await self.create_address(user_id=created_user.id)

        get_address_response = await self.address_service.soft_delete_address(address_id=created_address.id)
        gotten_address = get_address_response.data

        self.assertTrue(gotten_address)

# OBJECT FAKER FACTORIES using polyfactory
class CreateAddressDtoFactory(ModelFactory[CreateAddressDto]):
    __model__ = CreateAddressDto
    __check_model__ = False
    street_number = "23B"



from appodus_utils import Utils
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.method_trace_logger import method_trace_logger
from httpx import AsyncClient
from kink import di, inject
from starlette.requests import Request

from appodus_utils.domain.user.auth.models import SocialAuthOperationType, OAuthCallbackRequest, SocialLoginUserInfo
from appodus_utils.domain.user.auth.social_login.interface import ISocialAuthProvider
from appodus_utils.domain.user.auth.social_login.models import SocialAuthPlatform
from appodus_utils.domain.user.auth.social_login.providers.utils import OauthUtils

httpx_client: AsyncClient = di[AsyncClient]

@inject
@decorate_all_methods(method_trace_logger, exclude=['__init__'], exclude_startswith='_')
class FacebookAuthProvider(ISocialAuthProvider):
    def __init__(self):
        self._client_id = Utils.get_from_env_fail_if_not_exists("FACEBOOK_APP_ID")
        self._client_secret = Utils.get_from_env_fail_if_not_exists("FACEBOOK_APP_SECRET")
        self._auth_base_url = Utils.get_from_env_fail_if_not_exists("FACEBOOK_AUTH_BASE_URL")

    @property
    def platform(self):
        return SocialAuthPlatform.FACEBOOK

    async def initialize(self, operation_type: SocialAuthOperationType, request: Request) -> dict:
        scope = "openid email profile"
        redirect_url =  await OauthUtils.get_auth_redirect_url(self.platform, request)

        return  await OauthUtils.init_0auth(request, self._auth_base_url, self._client_id, redirect_url, scope, operation_type)

    async def verify(self, payload: OAuthCallbackRequest, request: Request) -> SocialLoginUserInfo:
        redirect_url =  await OauthUtils.get_auth_redirect_url(self.platform, request)

        # Exchange code for token
        token_response = await httpx_client.get(
            "https://graph.facebook.com/v22.0/oauth/access_token",
            params={
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "redirect_uri": redirect_url,
                "code": payload.code,
                "code_verifier": payload.code_verifier,
            }
        )
        token_response.raise_for_status()

        # Get user info
        user_response = await httpx_client.get(
            "https://graph.facebook.com/me",
            params={
                "fields": "id,email,first_name,last_name",
                "access_token": token_response.json()["access_token"]
            }
        )
        user_response.raise_for_status()
        user_info = user_response.json()

        return SocialLoginUserInfo(
            id=user_info["id"],
            email=user_info["email"],
            firstname=user_info["first_name"],
            lastname=user_info["last_name"],
            exp=0,
            operation_type=payload.operation_type
        )



from typing import Callable


def decorate_all_methods(
    decorator: Callable,
    exclude: list[str] = None,
    exclude_startswith: list[str] = None
):
    excluded = ['__init__', '__module__', '_session', '__dict__', '__weakref__', '__doc__']
    startswith_excluded = ["_", "__"]

    if exclude:
        excluded.extend(exclude)
    if exclude_startswith:
        startswith_excluded.extend(exclude_startswith)

    def decorate(cls):
        for attr_name, attr_value in cls.__dict__.items():
            if (
                callable(attr_value)
                and attr_name not in excluded
                and not attr_name.startswith(tuple(startswith_excluded))
            ):
                setattr(cls, attr_name, decorator(attr_value))
        return cls

    return decorate


# def decorate_all_methods(decorator: callable, exclude=None, exclude_startswith: str = None):
#     if exclude is None:
#         exclude = []
#
#     def decorate(cls):
#         for attr in cls.__dict__:
#             if (callable(getattr(cls, attr)) and (attr not in exclude) and
#                     (not exclude_startswith or not attr.startswith(exclude_startswith))):
#                 setattr(cls, attr, decorator(getattr(cls, attr)))
#         return cls
#
#     return decorate



from unittest import IsolatedAsyncioTestCase

from _pytest.monkeypatch import MonkeyPatch
from kink import di
from polyfactory.factories.pydantic_factory import ModelFactory


from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap

from appodus_utils.common.appodus_test_utils import TestUtils
from appodus_utils.db.session import close_db_engine
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.transactional import TransactionSessionPolicy, transactional
from appodus_utils.domain.user.device.models import Device, CreateDeviceDto, QueryDeviceDto, SearchDeviceDto
from appodus_utils.domain.user.device.service import DeviceService
from appodus_utils.domain.user.auth.models import LoginSuccessDto
from appodus_utils.domain.user.models import CreateUserDto, QueryUserDto, User
from appodus_utils.domain.user.service import UserService
from appodus_utils.integrations.messaging.models import PushToken
from tests.appodus_utils.domain.user.test_e2e_service import CreateUserDtoFactory
from tests.appodus_utils.test_utils import mock_active_auditor_service


@decorate_all_methods(transactional(session_policy=TransactionSessionPolicy.ALWAYS_NEW),
                      exclude=['__init__', 'asyncTearDown', 'asyncSetUp'], exclude_startswith='_')
class TestDeviceService(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.device_service: DeviceService = di[DeviceService]
        self.user_service: UserService = di[UserService]
        self.monkeypatch = MonkeyPatch()

    async def asyncTearDown(self):
        self.monkeypatch.undo()
        await self._truncate_tables()
        await close_db_engine()

    @staticmethod
    async def _truncate_tables():
        await TestUtils.truncate_entities([Device, User])

    async def create_user_and_mock_active_auditor_service(self) -> LoginSuccessDto:

        create_user_dto = CreateUserDtoFactory.build()
        # mock_active_auditor_service
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(create_user_dto))

        await self.user_service.store_verified_email_otp(
            email=create_user_dto.email,
            otp=create_user_dto.otp
        )

        created_user_response = await self.user_service.create_user(create_user_dto)

        # re-mock_active_auditor_service to include User ID
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(created_user_response))

        return created_user_response

    async def create_device(self, user_id: str) -> QueryDeviceDto:
        create_device_dto = CreateDeviceDtoFactory.build()
        create_device_dto.user_id = user_id

        created_device_response = await self.device_service.create_device(create_device_dto)
        created_device = created_device_response.data

        return created_device


    async def test_create_device(self):
        create_device_dto = CreateDeviceDtoFactory.build()

        created_device_response = await self.device_service.create_device(create_device_dto)
        created_device = created_device_response.data

        self.assertIsNotNone(created_device.id)

    async def test_get_device_page(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        count = 2
        for _ in range(count):
            await self.create_device(user_id=created_user.id)

        search_dto: SearchDeviceDto = SearchDeviceDto(
            page_size=10,
            page=0
        )

        device_page = await self.device_service.get_device_page(search_dto=search_dto)

        self.assertEqual(count, device_page.meta.count)

    async def test_update_device(self):
        created_user = await self.create_user_and_mock_active_auditor_service()
        created_device = await self.create_device(user_id=created_user.id)

        push_token: PushToken = PushToken(
            token="token",
        device_id="device_id"
        )

        update_device_response = await self.device_service.update_device_push_token(
            device_id=created_device.id,
            push_token=push_token
        )
        device_updated = update_device_response.data

        self.assertIsNotNone(device_updated)

# OBJECT FAKER FACTORIES using polyfactory
class CreateDeviceDtoFactory(ModelFactory[CreateDeviceDto]):
    __model__ = CreateDeviceDto
    __check_model__ = False
    street_number = "23B"



from datetime import timedelta
from unittest import IsolatedAsyncioTestCase

from _pytest.monkeypatch import MonkeyPatch
from kink import di
from polyfactory.factories.pydantic_factory import ModelFactory


from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap

from appodus_utils import Utils
from appodus_utils.common.appodus_test_utils import TestUtils
from appodus_utils.db.redis_utils import RedisUtils
from appodus_utils.db.session import close_db_engine
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.transactional import TransactionSessionPolicy, transactional
from appodus_utils.domain.user.auth.models import EmailValidationRequest, LoginSuccessDto
from appodus_utils.domain.user.models import User, QueryUserDto, CreateUserDto, SearchUserDto, UpdateNameDto, Gender
from appodus_utils.domain.user.service import UserService
from tests.appodus_utils.test_utils import mock_active_auditor_service


@decorate_all_methods(transactional(session_policy=TransactionSessionPolicy.ALWAYS_NEW),
                      exclude=['__init__', 'asyncTearDown', 'asyncSetUp'], exclude_startswith='_')
class TestUserService(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.user_service: UserService = di[UserService]
        self.monkeypatch = MonkeyPatch()

    async def asyncTearDown(self):
        self.monkeypatch.undo()
        await self._truncate_tables()
        await close_db_engine()

    @staticmethod
    async def _truncate_tables():
        await TestUtils.truncate_entities([User])

    async def create_user_and_mock_active_auditor_service(self, count: int = 1) -> LoginSuccessDto:

        create_user_dto = CreateUserDtoFactory.build()

        # Handle Unique Email
        create_user_dto.email = f"{count}_{create_user_dto.email}"

        # mock_active_auditor_service
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(create_user_dto))

        await self.user_service.store_verified_email_otp(
            email=create_user_dto.email,
            otp=create_user_dto.otp
        )

        created_user_response = await self.user_service.create_user(create_user_dto)

        # re-mock_active_auditor_service to include User ID
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(created_user_response))

        return created_user_response


    async def test_create_user(self):
        # PREPARE
        create_user_dto = CreateUserDtoFactory.build()
        # mock_active_auditor_service
        mock_active_auditor_service(monkeypatch=self.monkeypatch, user=QueryUserDto.model_copy(create_user_dto))
        await self.user_service.store_verified_email_otp(
            email=create_user_dto.email,
            otp=create_user_dto.otp
        )

        # ACT
        created_user_response = await self.user_service.create_user(create_user_dto)

        # ASSERT
        self.assertIsNotNone(created_user_response.id)

    async def test_get_user(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        get_user_response = await self.user_service.get_user(user_id=created_user.id)
        gotten_user = get_user_response.data

        # ASSERT
        self.assertIsNotNone(gotten_user.id)

    async def test_get_user_model(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        user = await self.user_service.get_user_model(user_id=created_user.id)

        # ASSERT
        self.assertIsNotNone(user.id)

    async def test_exists_by_email(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        user_exists = await self.user_service.exists_by_email(email=created_user.email)

        # ASSERT
        self.assertTrue(user_exists)

    async def test_get_by_email(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        get_user_response = await self.user_service.get_by_email(email=created_user.email)

        # ASSERT
        self.assertIsNotNone(get_user_response.id)

    async def test_get_user_page(self):
        # PREPARE
        count = 2
        for index in range(count):
            await self.create_user_and_mock_active_auditor_service(count=index)
        search_dto: SearchUserDto = SearchUserDto(
            page_size=10,
            page=0
        )

        # ACT
        user_page = await self.user_service.get_user_page(search_dto=search_dto)

        # ASSERT
        self.assertEqual(count, user_page.meta.count)

    async def test_update_user_change_password(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()
        new_password = "new_password"

        # ACT
        update_user_response = await self.user_service.update_user_change_password(
            new_password=new_password
        )
        updated_user = update_user_response.data

        # ASSERT
        self.assertEqual(2, updated_user.version)

    async def test_update_last_active(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        update_user_response = await self.user_service.update_last_active(
            user_id=created_user.id
        )
        updated_user = update_user_response.data

        # ASSERT
        self.assertEqual(2, updated_user.version)

    async def test_update_active_user_name(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()
        update_dto = UpdateNameDto(
            firstname="firstname",
            middle_name="middle_name",
            lastname="lastname"
        )

        # ACT
        update_user_response = await self.user_service.update_active_user_name(
            update_dto=update_dto
        )
        updated_user = update_user_response.data

        # ASSERT
        self.assertEqual(2, updated_user.version)

    async def test_soft_delete_user(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        get_user_response = await self.user_service.soft_delete_by_admin(user_id=created_user.id)
        user_deleted = get_user_response.data

        # ASSERT
        self.assertTrue(user_deleted)

    async def test_deactivate_active_user(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()
        reason = "deactivation reason"

        # ACT
        update_user_response = await self.user_service.deactivate_active_user(
            reason=reason
        )
        user_deactivated = update_user_response.data

        # ASSERT
        self.assertTrue(user_deactivated)

    async def test_activate_user(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()
        reason = "activation reason"

        # ACT
        update_user_response = await self.user_service.activate_user(
            user_id=created_user.id,
            reason=reason
        )
        user_activated = update_user_response.data

        # ASSERT
        self.assertTrue(user_activated)

    async def test_send_phone_validation_message(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        # ACT
        update_user_response = await self.user_service.send_phone_validation_message(
            phone_ext=created_user.phone_ext,
            phone=created_user.phone
        )
        user_activated = update_user_response.data

        # ASSERT
        self.assertTrue(user_activated)

    async def test_update_phone(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        otp="mock_otp"
        time_to_live = timedelta(seconds=30)
        phone_delimited = f'{created_user.phone_ext}:{created_user.phone}'

        store_key = f'{created_user.id}:{otp}'
        await RedisUtils.set_redis(store_key, phone_delimited, time_to_live)

        # ACT
        update_user_response = await self.user_service.update_phone(
            otp=otp
        )
        user_activated = update_user_response.data

        # ASSERT
        self.assertTrue(user_activated)

    async def test_send_email_validation_message(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()
        payload = EmailValidationRequest(
            email="kingsley.ezenwere@gmail.com"
        )
        is_a_new_user = False

        # ACT
        update_user_response = await self.user_service.send_email_validation_message(
            payload=payload,
            is_a_new_user=is_a_new_user
        )
        user_activated = update_user_response.data

        # ASSERT
        self.assertTrue(user_activated)

    async def test_store_verified_email_otp(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()
        otp="mock_otp"
        await self.user_service.store_verified_email_otp(
            email=created_user.email,
            otp=otp
        )

        # ACT
        await self.user_service._validate_email_verification_otp(
            email=created_user.email,
            otp=otp
        )

    async def test_validate_and_store_email_verification_otp(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        otp = Utils.get_otp_code()
        store_key = f'{created_user.email}:{otp}'

        time_to_live = timedelta(seconds=30)
        await RedisUtils.set_redis(store_key, created_user.email, time_to_live)

        # ACT
        store_validate_otp_response = await self.user_service.validate_and_store_email_verification_otp(
            email=created_user.email,
            otp=otp
        )
        validated_otp_stored = store_validate_otp_response.data

        # ASSERT
        self.assertTrue(validated_otp_stored)

    async def test_update_email(self):
        # PREPARE
        created_user = await self.create_user_and_mock_active_auditor_service()

        otp="mock_otp"

        await self.user_service.store_verified_email_otp(
            email=created_user.email,
            otp=otp
        )

        # ACT
        update_user_response = await self.user_service.update_email(
            email=created_user.email,
            otp=otp
        )
        user_activated = update_user_response.data

        # ASSERT
        self.assertTrue(user_activated)

    async def test_update_gender(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()

        # ACT
        update_user_response = await self.user_service.update_gender(
            gender=Gender.FEMALE
        )
        gender_updated = update_user_response.data

        # ASSERT
        self.assertTrue(gender_updated)

    async def test_update_dob(self):
        # PREPARE
        await self.create_user_and_mock_active_auditor_service()

        # ACT
        update_user_response = await self.user_service.update_dob(
            dob=Utils.datetime_now_to_db().date()
        )
        dob_updated = update_user_response.data

        # ASSERT
        self.assertTrue(dob_updated)


# OBJECT FAKER FACTORIES using polyfactory
class CreateUserDtoFactory(ModelFactory[CreateUserDto]):
    __model__ = CreateUserDto
    __check_model__ = False
    email="kingsley.ezenwere@gmail.com"
    phone="7039018727"
    phone_ext="234"


from datetime import datetime
from typing import Optional, Union

from pydantic import Field, HttpUrl, ConfigDict, field_serializer

from appodus_utils import Object
from appodus_utils.integrations.messaging.models import WebPushPayload, MessageChannel, MessageRequestRecipient, \
    EmailPayload, SmsPayload, WhatsappPayload, PushPayload, MessageStatus, MessageProviderName, MessagePriority, \
    MessageExtras


class MessageBaseDto(Object):
    pass

class UpsertMessageDto(MessageBaseDto):
    id: Optional[str] = Field(None, description="Unique message identifier")
    channel: MessageChannel = Field(..., description="Communication channel type")
    to: MessageRequestRecipient = Field(..., description="Recipient details")
    payload: Union[
        EmailPayload,
        SmsPayload,
        WhatsappPayload,
        PushPayload,
        WebPushPayload,
    ] = Field(..., description="Content to be sent")
    status: MessageStatus = Field(default=MessageStatus.PENDING, description="Current status of the message")
    provider: Optional[MessageProviderName] = Field(None, description="Message provider (e.g., Twilio, Mailjet)")
    provider_id: Optional[str] = Field(None, description="External ID returned by the provider")
    error: Optional[str] = Field(None, description="Last error message encountered, if any")
    retry_count: int = Field(default=0, ge=0, description="Number of retry attempts made so far")
    priority: MessagePriority = Field(default=MessagePriority.NORMAL, description="Delivery priority level")
    scheduled_at: Optional[datetime] = Field(None, description="Time at which message is scheduled to be sent")
    sent_at: Optional[datetime] = Field(None, description="Timestamp when the message was actually sent")
    delivered_at: Optional[datetime] = Field(None, description="Timestamp when the message was successfully delivered")
    extras: Optional[MessageExtras] = Field(None, description="Additional custom data or tracking metadata")
    callback_url: Optional[HttpUrl] = Field(None, description="Webhook URL for delivery status")
    # url_tags: Optional[str] = Field(default=None, description="The query tags passed for the CTA")
    # sandbox_mode: Optional[bool] = False

    # @model_validator(mode="after")
    # def validate_recipients(self) -> 'UpsertMessageDto':
    #     def normalize(val):
    #         if val is None:
    #             return []
    #         return val if isinstance(val, list) else [val]
    #
    #     all_recipients = normalize(self.to.recipient)
    #     cc_list = normalize(self.to.cc_recipient)
    #     bcc_list = normalize(self.to.bcc_recipient)
    #
    #     total = len(all_recipients) + len(cc_list) + len(bcc_list)
    #
    #     if self.channel in {MessageChannel.SMS, MessageChannel.WHATSAPP} and total > 1:
    #         raise ValueError(f"{self.channel.value.upper()} supports only one recipient")
    #
    #     if total > 1000:
    #         raise ValueError("Total recipients must not exceed 1000")
    #
    #     email_regex = r'^[^@]+@[^@]+\.[^@]+$'
    #     e164_regex = r'^\+[1-9]\d{1,14}$'
    #     wa_regex = r'^\d{1,15}$'
    #
    #     for recipient in all_recipients + cc_list + bcc_list:
    #         if self.channel == MessageChannel.EMAIL:
    #             if not re.fullmatch(email_regex, recipient):
    #                 raise ValueError(f"Invalid email address: {recipient}")
    #         elif self.channel == MessageChannel.SMS:
    #             if not re.fullmatch(e164_regex, recipient):
    #                 raise ValueError(f"Invalid SMS number (E.164): {recipient}")
    #         elif self.channel == MessageChannel.WHATSAPP:
    #             if not re.fullmatch(wa_regex, recipient):
    #                 raise ValueError(f"Invalid WhatsApp number: {recipient}")
    #         elif self.channel in {MessageChannel.PUSH, MessageChannel.WEB_PUSH}:
    #             if not isinstance(recipient, str) or len(recipient) > 256:
    #                 raise ValueError(f"Invalid device token: {recipient}")
    #
    #     if self.channel != MessageChannel.EMAIL and (cc_list or bcc_list):
    #         raise ValueError("CC and BCC are only supported for email channel")
    #
    #     return self


    @field_serializer("scheduled_at", "sent_at", "delivered_at")
    def serialize_datetimes(self, value: Optional[datetime], _info):
        return value.isoformat() if value else None

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "id": "msg_123",
                "channel": "email",
                "to": {
                    "user_id": "user_123",
                    "fullname": "Jane Doe",
                    "email": "jane.doe@example.com",
                    "phone": "+2348100000000",
                    "ios_push_token": [{"token": "abc123token", "device_id": "ios_device_1"}],
                    "cc_recipient": [
                        {"name": "Team Member", "email": "team@example.com"}
                    ]
                },
                "payload": {
                    "subject": "Welcome!",
                    "body": "<h1>Welcome</h1><p>Thank you for joining</p>"
                },
                "status": "pending",
                "priority": "normal",
                "extras": {
                    "campaign": "welcome_flow"
                }
            }
        }
    )



from enum import Enum


class AvailableTemplate(str, Enum):
    """
    Enum representing all available notification templates in the system,
    categorized by their purpose and including supported delivery channels.
    """

    # ============ ACCOUNT & SECURITY ============
    NEW_USER_WELCOME = "new_user_welcome"  # Email, WhatsApp, Mobile push
    NEW_BUYER_WELCOME = "new_buyer_welcome"  # Email, WhatsApp, Mobile push
    NEW_AGENT_WELCOME = "new_agent_welcome"  # Email, WhatsApp, Mobile push
    NEW_USER_EMAIL_VERIFICATION = "new_user_email_verification"  # Email
    EMAIL_VERIFICATION = "email_verification"  # Email
    PHONE_VERIFICATION = "phone_verification"  # SMS, WhatsApp
    NAME_UPDATE_SUCCESS= "name_update_success" # Email
    LOGIN_DIFF_DEVICE_SECURITY_ALERT = "login_diff_device_security_alert"  # SMS, Email, WhatsApp, Mobile push
    PASSWORD_UPDATE_SUCCESS = "password_update_success"  # Email
    PASSWORD_RESET_REQUEST = "password_reset_request"  # SMS, Email
    PASSWORD_RESET_SUCCESS = "password_reset_success"  # Email
    TWO_FA = "2fa"  # SMS, Email, WhatsApp
    ACCOUNT_DEACTIVATION = "account_deactivation"  # Email
    ACCOUNT_ACTIVATION = "account_activation"  # Email
    PROPERTY_LISTING_COMPLETION_REMINDER = "property_listing_completion_reminder"  # Email, WhatsApp, Mobile push

    # ============ LISTING MANAGEMENT ============
    PROPERTY_LISTED_SUCCESS = "property_listed_success"  # Email, WhatsApp
    PROPERTY_REJECTED_REASONS = "property_rejected_with_reasons"  # Email
    LISTING_PENDING_MODERATION = "listing_pending_moderation"  # Email
    LISTING_PERFORMANCE_SUMMARY = "listing_performance_summary"  # Email, WhatsApp
    PRICE_SUGGESTION_TRENDS = "price_suggestion_trends"  # Email, WhatsApp

    # ============ BUYER NOTIFICATIONS ============
    NEW_PROPERTY_MATCH_SAVED_SEARCH = "new_property_match_saved_search"  # Email, Mobile push, Web Push
    PRICE_DROP_SAVED_PROPERTY = "price_drop_saved_property"  # Email, Mobile push, Web Push
    SELLER_REPLIED_TO_INQUIRY = "seller_replied_to_inquiry"  # Email, WhatsApp, Mobile push, Web Push
    NEW_MESSAGE_FROM_SELLER_AGENT = "new_message_from_seller_agent"  # Email, WhatsApp, Mobile push
    INQUIRY_CONFIRMATION = "inquiry_confirmation"  # Email
    SCHEDULE_TOUR_CONFIRMATION = "schedule_tour_confirmation"  # SMS, Email, WhatsApp, Mobile push, Web Push
    TOUR_RESCHEDULED_CANCELLED = "tour_rescheduled_cancelled"  # SMS, Email, WhatsApp, Mobile push, Web Push

    # ============ REMINDERS & ALERTS ============
    BUYER_TOUR_REMINDER = "buyer_tour_reminder"  # SMS, Email, WhatsApp, Mobile push, Web Push
    INACTIVE_USER_REENGAGEMENT = "inactive_user_reengagement"  # Email, WhatsApp, Mobile push, Web Push

    # ============ AGENT ENGAGEMENT ============
    NEW_BUYER_LEAD_ALERT = "new_buyer_lead_alert"  # Email, WhatsApp, Mobile push
    BUYER_SCHEDULED_VISIT = "buyer_scheduled_visit"  # Email, WhatsApp, Mobile push
    BUYER_CANCELED_VISIT = "buyer_canceled_visit"  # Email, WhatsApp, Mobile push
    AGENT_PERFORMANCE_SUMMARY = "agent_performance_summary"  # Email
    NEW_REVIEW_FOR_AGENT = "new_review_for_agent"  # Email, Mobile push

    # ============ TRANSACTIONS ============
    OFFER_SUBMITTED_CONFIRMATION = "offer_submitted_confirmation"  # Email
    OFFER_RECEIVED_TO_SELLER = "offer_received_to_seller"  # Email, WhatsApp, Mobile push
    OFFER_ACCEPTED = "offer_accepted"  # Email, WhatsApp, Mobile push
    OFFER_REJECTED = "offer_rejected"  # Email
    COUNTER_OFFER_RECEIVED = "counter_offer_received"  # Email, WhatsApp, Mobile push
    ESCROW_INITIATED = "escrow_initiated"  # Email, WhatsApp
    ESCROW_STATUS_UPDATE = "escrow_status_update"  # Email, WhatsApp
    TRANSACTION_COMPLETED = "transaction_completed"  # Email, WhatsApp, Mobile push
    TRANSACTION_FAILED_DECLINED = "transaction_failed_declined"  # SMS, Email, WhatsApp

    # ============ POST-SALE & FOLLOW-UP ============
    POST_SALE_BUYER_FEEDBACK_REQUEST = "post_sale_buyer_feedback_request"  # Email, WhatsApp, Mobile push
    POST_SALE_SELLER_FEEDBACK_REQUEST = "post_sale_seller_feedback_request"  # Email, WhatsApp, Mobile push
    BUYER_REFERRAL_INCENTIVE = "buyer_referral_incentive"  # Email, WhatsApp, Mobile push
    SELLER_REFERRAL_INCENTIVE = "seller_referral_incentive"  # Email, WhatsApp, Mobile push

    # ============ TRUST & SAFETY ============
    ID_DOCUMENT_VERIFICATION_PENDING = "id_document_verification_pending"  # Email, Mobile push
    ID_DOCUMENT_REJECTED_ACCEPTED = "id_document_rejected_accepted"  # Email
    FRAUD_WARNING_ALERT = "fraud_warning_alert"  # SMS, Email, WhatsApp, Mobile push
    TERMS_POLICY_UPDATES = "terms_policy_updates"  # Email

    # ============ PRODUCT UPDATES ============
    NEW_FEATURE_ANNOUNCEMENT = "new_feature_announcement"  # Email, Mobile push, Web Push
    INVITE_TEST_NEW_FEATURE = "invite_test_new_feature"  # Email
    FEEDBACK_REQUEST_POST_TRANSACTION = "feedback_request_post_transaction"  # Email, WhatsApp, Mobile push
    SURVEY_PRODUCT_IMPROVEMENT = "survey_product_improvement"  # Email, WhatsApp

    # ============ MARKETING ============
    SELLER_TIPS_CONTENT = "seller_tips_content"  # Email, Mobile push, Web Push
    BUYER_TIPS_CONTENT = "buyer_tips_content"  # Email, Mobile push, Web Push
    REFERRAL_PROGRAM_INVITATION = "referral_program_invitation"  # Email, WhatsApp, Mobile push
    HOLIDAY_FESTIVE_GREETING = "holiday_festive_greeting"  # Email, WhatsApp
    LIMITED_TIME_PROMO_DISCOUNT = "limited_time_promo_discount"  # Email, WhatsApp, Mobile push, Web Push



from logging import Logger

from kink import inject, di

from appodus_utils import Page
from appodus_utils.db.models import SuccessResponse
from appodus_utils.decorators.decorate_all_methods import decorate_all_methods
from appodus_utils.decorators.method_trace_logger import method_trace_logger
from appodus_utils.decorators.transactional import transactional
from appodus_utils.domain.user.address.models import CreateAddressDto, QueryAddressDto, SearchAddressDto, \
    UpdateAddressDto, _UpdateAddressDto
from appodus_utils.domain.user.address.repo import AddressRepo
from appodus_utils.domain.user.address.validator import AddressValidator
from appodus_utils.domain.user.auth.active_auditor.service import ActiveAuditorService

logger: Logger = di['logger']


@inject
@decorate_all_methods(transactional(), exclude=['__init__'], exclude_startswith='_')
@decorate_all_methods(method_trace_logger, exclude=['__init__'], exclude_startswith='_')
class AddressService:
    def __init__(self, address_repo: AddressRepo,
                 address_validator: AddressValidator,
                 active_auditor_service: ActiveAuditorService):
        self._address_repo = address_repo
        self._address_validator = address_validator
        self._active_auditor_service = active_auditor_service

    async def create_address(self, create_dto: CreateAddressDto) -> SuccessResponse[QueryAddressDto]:
        return await self._address_repo.create(create_dto)

    async def get_address(self, address_id: str) -> SuccessResponse[QueryAddressDto]:
        await self._address_validator.should_exist_by_id(address_id)
        return await self._address_repo.get(address_id)

    async def get_address_page(self, search_dto: SearchAddressDto) -> Page[QueryAddressDto]:
        return await self._address_repo.get_page(search_dto=search_dto)

    async def get_address_page_for_active_user(self) -> Page[QueryAddressDto]:
        _, authorizer = await self._active_auditor_service.get_combined_authorizer_from_context()
        user_id = authorizer.get_jwt_subject()
        search_dto: SearchAddressDto = SearchAddressDto(user_id=user_id)
        return await self._address_repo.get_page(search_dto=search_dto)

    async def update_address(self, address_id: str, update_dto: UpdateAddressDto) -> SuccessResponse[QueryAddressDto]:
        await self._address_validator.should_exist_by_id(address_id)
        return await self._address_repo.update(address_id, update_dto)

    async def verify_address(self, address_id: str) -> SuccessResponse[QueryAddressDto]:
        await self._address_validator.should_exist_by_id(address_id)
        update_dto = _UpdateAddressDto(verified=True)
        return await self._address_repo.update(address_id, update_dto.model_dump(exclude_none=True))

    async def soft_delete_address(self, address_id: str) -> SuccessResponse[bool]:
        await self._address_validator.should_exist_by_id(address_id)
        await self._address_repo.soft_delete(address_id)

        return SuccessResponse(
                    data=True
                )



import json
from contextvars import Token
from datetime import timedelta, datetime
from logging import Logger
from typing import Any, Union, Dict, Optional, List

from fastapi import BackgroundTasks
from fastapi.encoders import jsonable_encoder
from kink import inject, di
from libre_fastapi_jwt import AuthJWT

from appodus_utils import Utils
from appodus_utils.db.redis_utils import RedisUtils
from appodus_utils.domain.user.auth.active_auditor.context import ActiveAuditorContext
from appodus_utils.domain.user.auth.active_auditor.models import ActiveAuditor
from appodus_utils.domain.user.models import UserPersona
from appodus_utils.exception.exceptions import NoActiveSessionException

logger: Logger = di['logger']


@inject
class ActiveAuditorService:

    @staticmethod
    async def login(user_id: str, auth_json: Any) -> ActiveAuditor:
        access_token_expire_seconds = Utils.get_from_env_fail_if_not_exists("ACCESS_TOKEN_EXPIRE_SECONDS")
        time_to_live = timedelta(seconds=int(access_token_expire_seconds))
        active_auditor = ActiveAuditor(**auth_json)
        active_auditor_json = jsonable_encoder(active_auditor)
        active_auditor_str = json.dumps(active_auditor_json)
        await RedisUtils.set_redis(user_id, active_auditor_str, time_to_live)

        return active_auditor

    async def init_context(self, background_tasks: BackgroundTasks, authorizer: AuthJWT) -> tuple[
        Optional[Token[BackgroundTasks]], Optional[Token[AuthJWT]], Optional[Token[ActiveAuditor]]]:
        background_tasks_context_token: Token[BackgroundTasks] = ActiveAuditorContext.set_background_tasks_context(
            background_tasks)
        authorizer_context_token: Token[AuthJWT] = ActiveAuditorContext.set_authorizer_context(authorizer)

        user_id = await self.get_active_user_id(authorizer, False)
        if not user_id:
            return background_tasks_context_token, authorizer_context_token, None

        active_auditor_data = await RedisUtils.get_redis(user_id)
        if not active_auditor_data:
            raise NoActiveSessionException()
        active_auditor_data = json.loads(active_auditor_data)

        active_auditor = ActiveAuditor(**active_auditor_data)
        active_auditor_context_token: Token[ActiveAuditor] = ActiveAuditorContext.set_active_auditor_context(
            active_auditor)

        return background_tasks_context_token, authorizer_context_token, active_auditor_context_token

    @staticmethod
    async def reset_context(background_tasks_context_token: Optional[Token[BackgroundTasks]],
                            authorizer_context_token: Optional[Token[AuthJWT]],
                            active_auditor_context_token: Optional[Token[ActiveAuditor]]):
        if background_tasks_context_token:
            ActiveAuditorContext.reset_background_tasks_context(background_tasks_context_token)
        if authorizer_context_token:
            ActiveAuditorContext.reset_authorizer_context(authorizer_context_token)
        if active_auditor_context_token:
            ActiveAuditorContext.reset_active_auditor_context(active_auditor_context_token)

    @staticmethod
    async def get_background_tasks_from_context() -> BackgroundTasks:
        background_tasks: BackgroundTasks = ActiveAuditorContext.get_background_tasks_from_context()
        return background_tasks

    @staticmethod
    async def get_combined_authorizer_from_context() -> tuple[ActiveAuditor, AuthJWT]:
        active_auditor: ActiveAuditor = ActiveAuditorContext.get_active_auditor_from_context()
        authorizer: AuthJWT = ActiveAuditorContext.get_authorizer_from_context()
        return active_auditor, authorizer


    async def get_active_user_id(self, authorizer: AuthJWT = None, jwt_required: bool = True) -> Optional[str]:
        try:
            if not authorizer:
                authorizer = await self.get_authorizer_from_context()
            authorizer.jwt_required()
            return authorizer.get_jwt_subject()
        except Exception as e:
            if jwt_required:
                raise
            return None

    @staticmethod
    async def get_authorizer_from_context() -> AuthJWT:
        return ActiveAuditorContext.get_authorizer_from_context()

    async def get_user_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.id

        return None

    async def get_dob(self) -> Optional[datetime]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.dob

        return None

    async def get_firstname(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.firstname

        return None

    async def get_lastname(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.lastname

        return None

    async def get_fullname(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.fullname

        return None

    async def get_phone(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.phone

        return None

    async def get_email(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.email

        return None

    async def get_user_type(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.user_type

        return None

    async def get_personas(self) -> Optional[List[UserPersona]]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.personas

        return None

    async def get_last_active_date(self) -> Optional[datetime]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.last_active_date

        return None

    async def has_profile_picture(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.has_profile_picture

        return None

    async def has_selfie_picture(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.has_selfie_picture

        return None

    async def is_phone_validated(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.phone_validated

        return None

    async def is_email_validated(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.email_validated

        return None

    async def get_languages(self) -> Optional[List[str]]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.languages

        return None

    async def is_bvn_validated(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.bvn_validated

        return None

    async def is_identity_validated(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.identity_validated

        return None

    async def is_address_validated(self) -> Optional[bool]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.address_validated

        return None

    async def get_profile_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.profile_id

        return None

    async def get_wallet_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.wallet_id

        return None

    async def get_escrow_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.escrow_id

        return None

    async def get_lien_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.lien_id

        return None

    async def get_inbox_id(self) -> Optional[str]:
        user, _ = await self.get_combined_authorizer_from_context()
        if user:
            return user.inbox_id

        return None

    async def update(self, obj_in: Union[Any, Dict[str, Any]]) -> ActiveAuditor:
        active_auditor, _ = await self.get_combined_authorizer_from_context()
        active_auditor_data = jsonable_encoder(active_auditor)
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.model_dump(exclude_unset=True, exclude_none=True)

        for field in active_auditor_data:
            if field in update_data:
                try:
                    setattr(active_auditor, field, update_data[field])
                except AttributeError as e:
                    pass

        active_auditor_data = jsonable_encoder(active_auditor)
        return await self.login(active_auditor.id, active_auditor_data)



from typing import List

from kink import di

from appodus_utils.integrations.messaging.models import MessageRequest, BatchResult
from appodus_utils.sdk.appodus_sdk.services.messages.messaging_service import MessagingService
from appodus_utils.sdk.appodus_sdk.utils import AppodusClientUtils


class MessageClient:
    def __init__(self, message_manager_url: str, client_utils: AppodusClientUtils):
        self._client_utils = client_utils
        self._message_manager_url = message_manager_url
        self.messaging_service = MessagingService(
            message_manager_url=message_manager_url,
            client_utils=client_utils
        )

        # Provide for further injection by builder style message sender
        di[MessagingService] = lambda _di: self.messaging_service

    async def send_bulk(self, message_requests: List[MessageRequest]) -> BatchResult:
        return await self.messaging_service.send_bulk(message_requests=message_requests)

    async def send_message(self, message_request: MessageRequest) -> BatchResult:
        return await self.messaging_service.send_message(message_request=message_request)



from typing import List

from appodus_utils import Object
from appodus_utils.integrations.document_sign.models import Signer


class GenerateAndSendSignDto(Object):
    template_id: str
    template_variables: dict[str, str]
    contract_id: str
    signers: List[Signer]
    signing_client_name: str



import ast
import json
import re
from operator import attrgetter
from typing import Dict, List, Optional, Set, ClassVar, TypeVar, Type, Tuple

import cachetools

from appodus_utils.exception.exceptions import ValidationException
from pydantic import Field, PrivateAttr, field_validator, ConfigDict

from appodus_utils import Object, Utils
from appodus_utils.integrations.messaging.models import MessagePriority

T = TypeVar("T", bound="MessagingConfig")

class MessagingConfig(Object):
    """Immutable messaging configuration with validated fields."""

    _subject_cache = cachetools.LRUCache(maxsize=54)

    # Public fields (immutable by convention)
    from_email: str
    from_name: str = Field(..., min_length=1, max_length=100)
    sms_ttl: int = Field(..., gt=60, le=86400)
    sms_sender_id: str = Field(..., min_length=1, max_length=11)
    headers: Dict[str, str] = Field(default_factory=dict)
    priority: MessagePriority = MessagePriority.NORMAL
    sandbox_mode: bool = False
    brand_name: str
    otp_token_expire_mins: str
    categories: Optional[List[str]] = Field(default_factory=list)
    subjects: Dict[str, str] = Field(
        default_factory=dict,
        description="Mapping of subject keys to email subject templates"
    )

    # Private attributes
    _normalized_subjects: Tuple[Tuple[str, str], ...] = PrivateAttr()

    def normalize_subjects(self):
        normalized_subjects = {
            k.replace('_subject', '').lower(): v
            for k, v in self.subjects.items()
        }
        self._normalized_subjects = tuple(sorted(normalized_subjects.items()))

    # Class-level constants
    _required_subjects: ClassVar[Set[str]] = {
        'otp',
        'account_activation',
        'account_deactivation',
        'email_verification',
        'login_diff_device_security_alert',
        'name_update_success',
        'new_feature_announcement',
        'new_user_email_verification',
        'new_user_welcome',
        'password_reset_request',
        'password_update_success',
        'phone_verification',
    }

    # Pydantic v2 config
    model_config = ConfigDict(
        frozen=True,  # Makes instances immutable
        extra='forbid',  # Prevents extra fields
        str_strip_whitespace=True,  # Auto-trim strings
    )

    @field_validator('from_name')
    @classmethod
    def validate_from_name(cls, v: str) -> str:
        if not re.match(r'^[\w\s\-\.]+$', v):
            raise ValueError("From name contains invalid characters")
        return v

    @field_validator('sms_sender_id')
    @classmethod
    def validate_sms_id(cls, v: str) -> str:
        if not re.match(r'^[a-zA-Z0-9\s]+$', v):
            raise ValidationException(message="SMS SENDER ID allows only alphanumeric chars and spaces")
        return v

    @field_validator('subjects')
    @classmethod
    def validate_subjects(cls, v: Dict[str, str]) -> Dict[str, str]:
        normalized_keys = {k.replace('_subject', '').lower() for k in v}
        missing = cls._required_subjects - normalized_keys
        if missing:
            raise ValueError(f"Missing required subjects: {', '.join(sorted(missing))}")

        for key, value in v.items():
            if len(value) > 150:
                raise ValueError(f"Subject '{key}' exceeds 150 character limit")
        return v

    # --- Init hook ---
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.normalize_subjects()

    # --- Cached subject lookup ---
    @cachetools.cachedmethod(attrgetter('_subject_cache'), key=lambda self, key: key)
    def get_subject(self, key: str) -> str:
        """O(1) subject lookup."""
        normalized = key.replace('_subject', '').lower()
        if subject := dict(self._normalized_subjects).get(normalized):
            return subject
        raise KeyError(f"Invalid subject key: {key}")

    # --- Constructor from app settings ---
    @classmethod
    def from_settings(cls: Type[T]) -> T:
        """Preferred constructor from settings."""
        EMAIL_FROM_ADDRESS = Utils.get_from_env_fail_if_not_exists("EMAIL_FROM_ADDRESS")
        EMAIL_FROM_NAME = Utils.get_from_env_fail_if_not_exists("EMAIL_FROM_NAME")
        SMS_TTL = Utils.get_from_env_fail_if_not_exists("SMS_TTL")
        SMS_SENDER_ID = Utils.get_from_env_fail_if_not_exists("SMS_SENDER_ID")
        MESSAGING_HEADERS = Utils.get_from_env("MESSAGING_HEADERS")
        MESSAGING_PRIORITY = Utils.get_from_env("MESSAGING_PRIORITY")
        MESSAGING_SANDBOX_MODE = Utils.get_from_env("MESSAGING_SANDBOX_MODE")
        MESSAGING_CATEGORIES = Utils.get_from_env("MESSAGING_CATEGORIES")
        EMAIL_SUBJECTS = Utils.get_from_env("EMAIL_SUBJECTS")
        MESSAGING_BRAND_NAME = Utils.get_from_env_fail_if_not_exists("MESSAGING_BRAND_NAME")
        OTP_TOKEN_EXPIRE_SECONDS = Utils.get_from_env_fail_if_not_exists("OTP_TOKEN_EXPIRE_SECONDS")


        return cls(
            from_email=EMAIL_FROM_ADDRESS,
            from_name=EMAIL_FROM_NAME,
            sms_ttl=int(SMS_TTL) if SMS_TTL else 25000,
            sms_sender_id=SMS_SENDER_ID,
            headers=ast.literal_eval(MESSAGING_HEADERS) if MESSAGING_HEADERS else {},
            priority=MESSAGING_PRIORITY or 2,
            sandbox_mode=bool(MESSAGING_SANDBOX_MODE) if MESSAGING_SANDBOX_MODE else False,
            categories=json.loads(MESSAGING_CATEGORIES) if MESSAGING_CATEGORIES else [],
            subjects=ast.literal_eval(EMAIL_SUBJECTS) if EMAIL_SUBJECTS else {},
            brand_name = MESSAGING_BRAND_NAME,
            otp_token_expire_mins = f"{(int(OTP_TOKEN_EXPIRE_SECONDS) / 60)}" if OTP_TOKEN_EXPIRE_SECONDS else "5"

        )



from logging import Logger
from typing import Any, Dict, List

from fastapi import BackgroundTasks
from kink import inject, di

from main.app.config.settings import settings
from main.app.domain.message.message_payload_builder import MessageRecipientBuilder
from main.app.domain.user.auth.active_auditor.service import ActiveAuditorService
from main.app.integrations.messaging.channel_sender import MessageDispatcher
from main.app.integrations.messaging.models import (
    MessageChannel, MultiChannelMessageRequest,
    MessageRequestRecipient, MessageCategory, MessageRecipientUserId, MessageContextModule
)
from main.app.integrations.messaging.templating.models import AvailableTemplate

logger: Logger = di["logger"]


@inject
class ListingManagementMessages(BaseMessageSender):
    """Handles all property listing related notifications"""

    async def send_property_listed_success_message(self, recipient_user_id: MessageRecipientUserId,
                                                   context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.PROPERTY_LISTED_SUCCESS,
            context_modules=context_modules,
            category=MessageCategory.LISTING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )

    async def send_property_rejected_message(self, recipient_user_id: MessageRecipientUserId,
                                             context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.PROPERTY_REJECTED_REASONS,
            context_modules=context_modules,
            category=MessageCategory.LISTING,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_listing_pending_moderation_message(self, recipient_user_id: MessageRecipientUserId,
                                                      context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.LISTING_PENDING_MODERATION,
            context_modules=context_modules,
            category=MessageCategory.LISTING,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_listing_performance_summary(self, recipient_user_id: MessageRecipientUserId,
                                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.LISTING_PERFORMANCE_SUMMARY,
            context_modules=context_modules,
            category=MessageCategory.ANALYTICS,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )

    async def send_price_suggestion_trends(self, recipient_user_id: MessageRecipientUserId,
                                           context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.PRICE_SUGGESTION_TRENDS,
            context_modules=context_modules,
            category=MessageCategory.ANALYTICS,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )


@inject
class BuyerNotificationMessages(BaseMessageSender):
    """Handles all buyer-related notifications"""

    async def send_new_property_match(self, recipient_user_id: MessageRecipientUserId,
                                      context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.NEW_PROPERTY_MATCH_SAVED_SEARCH,
            context_modules=context_modules,
            category=MessageCategory.ALERT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_price_drop_alert(self, recipient_user_id: MessageRecipientUserId,
                                    context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.PRICE_DROP_SAVED_PROPERTY,
            context_modules=context_modules,
            category=MessageCategory.ALERT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_seller_reply_notification(self, recipient_user_id: MessageRecipientUserId,
                                             context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.SELLER_REPLIED_TO_INQUIRY,
            context_modules=context_modules,
            category=MessageCategory.MESSAGE,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_new_message_from_agent(self, recipient_user_id: MessageRecipientUserId,
                                          context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.NEW_MESSAGE_FROM_SELLER_AGENT,
            context_modules=context_modules,
            category=MessageCategory.MESSAGE,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_inquiry_confirmation(self, recipient_user_id: MessageRecipientUserId,
                                        context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.INQUIRY_CONFIRMATION,
            context_modules=context_modules,
            category=MessageCategory.CONFIRMATION,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_tour_confirmation(self, recipient_user_id: MessageRecipientUserId,
                                     context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.SCHEDULE_TOUR_CONFIRMATION,
            context_modules=context_modules,
            category=MessageCategory.CONFIRMATION,
            default_channels=[
                MessageChannel.SMS,
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_tour_rescheduled_cancelled(self, recipient_user_id: MessageRecipientUserId,
                                              context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.TOUR_RESCHEDULED_CANCELLED,
            context_modules=context_modules,
            category=MessageCategory.UPDATE,
            default_channels=[
                MessageChannel.SMS,
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )


@inject
class ReminderAlertMessages(BaseMessageSender):
    """Handles all reminder and alert notifications"""

    async def send_tour_reminder(self, recipient_user_id: MessageRecipientUserId,
                                 context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.BUYER_TOUR_REMINDER,
            context_modules=context_modules,
            category=MessageCategory.REMINDER,
            default_channels=[
                MessageChannel.SMS,
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_inactive_user_reengagement(self, recipient_user_id: MessageRecipientUserId,
                                              context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.INACTIVE_USER_REENGAGEMENT,
            context_modules=context_modules,
            category=MessageCategory.REENGAGEMENT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )


@inject
class AgentEngagementMessages(BaseMessageSender):
    """Handles all agent engagement notifications"""

    async def send_new_buyer_lead_alert(self, recipient_user_id: MessageRecipientUserId,
                                        context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.NEW_BUYER_LEAD_ALERT,
            context_modules=context_modules,
            category=MessageCategory.LEAD,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_buyer_scheduled_visit(self, recipient_user_id: MessageRecipientUserId,
                                         context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.BUYER_SCHEDULED_VISIT,
            context_modules=context_modules,
            category=MessageCategory.APPOINTMENT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_buyer_canceled_visit(self, recipient_user_id: MessageRecipientUserId,
                                        context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.BUYER_CANCELED_VISIT,
            context_modules=context_modules,
            category=MessageCategory.APPOINTMENT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_agent_performance_summary(self, recipient_user_id: MessageRecipientUserId,
                                             context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.AGENT_PERFORMANCE_SUMMARY,
            context_modules=context_modules,
            category=MessageCategory.ANALYTICS,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_new_agent_review_notification(self, recipient_user_id: MessageRecipientUserId,
                                                 context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.NEW_REVIEW_FOR_AGENT,
            context_modules=context_modules,
            category=MessageCategory.REVIEW,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH
            ]
        )


@inject
class TransactionMessages(BaseMessageSender):
    """Handles all transaction-related notifications"""

    async def send_offer_submitted_confirmation(self, recipient_user_id: MessageRecipientUserId,
                                                context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.OFFER_SUBMITTED_CONFIRMATION,
            context_modules=context_modules,
            category=MessageCategory.CONFIRMATION,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_offer_received_notification(self, recipient_user_id: MessageRecipientUserId,
                                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.OFFER_RECEIVED_TO_SELLER,
            context_modules=context_modules,
            category=MessageCategory.OFFER,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_offer_accepted_notification(self, recipient_user_id: MessageRecipientUserId,
                                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.OFFER_ACCEPTED,
            context_modules=context_modules,
            category=MessageCategory.OFFER,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_offer_rejected_notification(self, recipient_user_id: MessageRecipientUserId,
                                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.OFFER_REJECTED,
            context_modules=context_modules,
            category=MessageCategory.OFFER,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_counter_offer_received(self, recipient_user_id: MessageRecipientUserId,
                                          context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.COUNTER_OFFER_RECEIVED,
            context_modules=context_modules,
            category=MessageCategory.OFFER,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_escrow_initiated_notification(self, recipient_user_id: MessageRecipientUserId,
                                                 context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.ESCROW_INITIATED,
            context_modules=context_modules,
            category=MessageCategory.TRANSACTION,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )

    async def send_escrow_status_update(self, recipient_user_id: MessageRecipientUserId,
                                        context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.ESCROW_STATUS_UPDATE,
            context_modules=context_modules,
            category=MessageCategory.TRANSACTION,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )

    async def send_transaction_completed_notification(self, recipient_user_id: MessageRecipientUserId,
                                                      context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.TRANSACTION_COMPLETED,
            context_modules=context_modules,
            category=MessageCategory.TRANSACTION,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_transaction_failed_notification(self, recipient_user_id: MessageRecipientUserId,
                                                   context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.TRANSACTION_FAILED_DECLINED,
            context_modules=context_modules,
            category=MessageCategory.TRANSACTION,
            default_channels=[
                MessageChannel.SMS,
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )


@inject
class PostSaleMessages(BaseMessageSender):
    """Handles all post-sale follow-up messages"""

    async def send_buyer_feedback_request(self, recipient_user_id: MessageRecipientUserId,
                                          context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.POST_SALE_BUYER_FEEDBACK_REQUEST,
            context_modules=context_modules,
            category=MessageCategory.FEEDBACK,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_seller_feedback_request(self, recipient_user_id: MessageRecipientUserId,
                                           context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.POST_SALE_SELLER_FEEDBACK_REQUEST,
            context_modules=context_modules,
            category=MessageCategory.FEEDBACK,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_buyer_referral_incentive(self, recipient_user_id: MessageRecipientUserId,
                                            context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.BUYER_REFERRAL_INCENTIVE,
            context_modules=context_modules,
            category=MessageCategory.REFERRAL,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_seller_referral_incentive(self, recipient_user_id: MessageRecipientUserId,
                                             context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.SELLER_REFERRAL_INCENTIVE,
            context_modules=context_modules,
            category=MessageCategory.REFERRAL,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )


@inject
class TrustSafetyMessages(BaseMessageSender):
    """Handles all trust and safety related notifications"""

    async def send_id_verification_pending(self, recipient_user_id: MessageRecipientUserId,
                                           context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.ID_DOCUMENT_VERIFICATION_PENDING,
            context_modules=context_modules,
            category=MessageCategory.VERIFICATION,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH
            ]
        )

    async def send_id_verification_result(self, recipient_user_id: MessageRecipientUserId,
                                          context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.ID_DOCUMENT_REJECTED_ACCEPTED,
            context_modules=context_modules,
            category=MessageCategory.VERIFICATION,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_fraud_warning_alert(self, recipient_user_id: MessageRecipientUserId,
                                       context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.FRAUD_WARNING_ALERT,
            context_modules=context_modules,
            category=MessageCategory.SECURITY,
            default_channels=[
                MessageChannel.SMS,
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_terms_policy_update(self, recipient_user_id: MessageRecipientUserId,
                                       context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.TERMS_POLICY_UPDATES,
            context_modules=context_modules,
            category=MessageCategory.ADMIN,
            default_channels=[MessageChannel.EMAIL]
        )


@inject
class ProductUpdateMessages(BaseMessageSender):
    """Handles all product update and feature announcements"""

    async def send_new_feature_announcement(self, recipient_user_id: MessageRecipientUserId,
                                            context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.NEW_FEATURE_ANNOUNCEMENT,
            context_modules=context_modules,
            category=MessageCategory.PRODUCT_IMPROVEMENT,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_feature_test_invitation(self, recipient_user_id: MessageRecipientUserId,
                                           context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.INVITE_TEST_NEW_FEATURE,
            context_modules=context_modules,
            category=MessageCategory.PRODUCT_IMPROVEMENT,
            default_channels=[MessageChannel.EMAIL]
        )

    async def send_post_transaction_feedback_request(self, recipient_user_id: MessageRecipientUserId,
                                                     context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.FEEDBACK_REQUEST_POST_TRANSACTION,
            context_modules=context_modules,
            category=MessageCategory.FEEDBACK,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_product_improvement_survey(self, recipient_user_id: MessageRecipientUserId,
                                              context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.SURVEY_PRODUCT_IMPROVEMENT,
            context_modules=context_modules,
            category=MessageCategory.FEEDBACK,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )


@inject
class MarketingMessages(BaseMessageSender):
    """Handles all marketing and promotional messages"""

    async def send_seller_tips(self, recipient_user_id: MessageRecipientUserId,
                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.SELLER_TIPS_CONTENT,
            context_modules=context_modules,
            category=MessageCategory.MARKETING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_buyer_tips(self, recipient_user_id: MessageRecipientUserId,
                              context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.BUYER_TIPS_CONTENT,
            context_modules=context_modules,
            category=MessageCategory.MARKETING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )

    async def send_referral_program_invitation(self, recipient_user_id: MessageRecipientUserId,
                                               context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.REFERRAL_PROGRAM_INVITATION,
            context_modules=context_modules,
            category=MessageCategory.MARKETING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH
            ]
        )

    async def send_holiday_greeting(self, recipient_user_id: MessageRecipientUserId,
                                    context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.HOLIDAY_FESTIVE_GREETING,
            context_modules=context_modules,
            category=MessageCategory.MARKETING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP
            ]
        )

    async def send_limited_time_promo(self, recipient_user_id: MessageRecipientUserId,
                                      context_modules:List[MessageContextModule]):
        await self._send_message(
            recipient_user_id=recipient_user_id,
            template=AvailableTemplate.LIMITED_TIME_PROMO_DISCOUNT,
            context_modules=context_modules,
            category=MessageCategory.MARKETING,
            default_channels=[
                MessageChannel.EMAIL,
                MessageChannel.WHATSAPP,
                MessageChannel.PUSH,
                MessageChannel.WEB_PUSH
            ]
        )



from kink import inject

from appodus_utils.domain.user.repo import UserRepo
from appodus_utils.exception.exceptions import ResourceNotFoundException, ResourceConflictException


@inject
class UserValidator:
    def __init__(self, user_repo: UserRepo):
        self._user_repo = user_repo

    async def should_exist_by_id(self, _id: str):
        if not (await self._user_repo.exists_by_id(_id)):
            raise ResourceNotFoundException("User")

    async def should_not_exist_by_email(self, email: str):
        if await self._user_repo.exists_by_email(email):
            raise ResourceConflictException(resource="User", message="A user with this email already exists")



from typing import Dict, Any

from kink import inject

from appodus_utils.exception.exceptions import TemplateRenderingException
from appodus_utils.integrations.messaging.models import MessageChannel
from appodus_utils.sdk.appodus_sdk.services.messages.templating.factory import TemplateEngineProviderFactory
from appodus_utils.sdk.appodus_sdk.services.messages.templating.interface import ITemplateEngineProvider


@inject
class TemplateService:
    def __init__(self, template_engine_provider_factory: TemplateEngineProviderFactory):
        self._template_engine_provider: ITemplateEngineProvider = template_engine_provider_factory.get_active_provider()

    async def render_message(
            self,
            channel: MessageChannel,
            template_name: str,
            context: Dict[str, Any]
    ) -> str:
        full_template_path = f"{channel.value}/{template_name}.{self._template_engine_provider.template_extension}"

        try:
            if not self._template_engine_provider.supports_template(full_template_path):
                raise TemplateRenderingException(f"Template not found: {full_template_path}")

            return self._template_engine_provider.render(full_template_path, context)
        except ValueError as e:
            raise TemplateRenderingException(str(e))



import os
from typing import List

from kink import inject

from appodus_utils import Utils
from appodus_utils.common.utils_settings import utils_settings
from appodus_utils.config.bootstrap import BaseDiBootstrap
from appodus_utils.sdk.appodus_sdk.services.messages.templating.interface import ITemplateEngineProvider

BaseDiBootstrap.register_all_subclasses(ITemplateEngineProvider)


@inject
class TemplateEngineProviderFactory:
    def __init__(self, providers: List[ITemplateEngineProvider]):
        self.ACTIVE_TEMPLATING_ENGINE = utils_settings.ACTIVE_TEMPLATING_ENGINE

        self._providers = providers
        self._factory = {}
        self._init_factory()

    def _init_factory(self):
        for provider in self._providers:
            self._factory[provider.template_engine] = provider

    def get_active_provider(self) -> ITemplateEngineProvider:
        return self._factory.get(self.ACTIVE_TEMPLATING_ENGINE)
